# Copyright (C) 2022 Alexandre Mitsuru Kaihara
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import subprocess
import hashlib
from configparser import ConfigParser
import json
from .exceptions import *
from .constants import *


# Just to enable the declaration of Type in methods
class Node:
    pass


# Brief: This is a super class that define methods common for network nodes
class Node:
    # Brief: Constructor of Node super class
    # Params:
    #   String containerName: Name of the container
    # Return:
    #   None
    def __init__(self, nodeName: str) -> None:
        self.__createTmpFolder()
        self.__nodeName = nodeName
        self.memory = ''
        self.cpu = ''

    def __createTmpFolder(self) -> None:
        subprocess.run("mkdir -p /tmp/lft/", shell=True)

    # OBS: Create nodes with short name lenght due to a restriction on a iproute2 to define and create interfaces.
    # Brief: Instantiate the container
    # Params:
    #   String dockerImage: Name of the container of a local image or in a Docker Hub repository
    #   String DockerCommand: String to be used to instantiate the container instead of the standard command
    #   String memory: It is the amount of memory to be allocated to the container (e.g. "512m", which is 512 MB)
    #   String cpus: It is the amount of cpu dedicated to the container, can be a fractional value such as "0.5"
    # Return:
    #   None
    def instantiate(self, dockerImage="alexandremitsurukaihara/lst2.0:host", dockerCommand='', dns='8.8.8.8', memory='', cpus='', runCommand='') -> None:
        command = []
        
        def addDockerRun():
            command.append(DOCKER_RUN)

        def addRunOptions():
            command.append("-d")

        def addNetwork():
            command.append(NETWORK + "=none")

        def addContainerName():
            command.append(NAME + '=' + self.getNodeName())

        def addPrivileged():
            command.append(PRIVILEGED)

        def addDNS(dns):
            command.append(DNS + '=' + dns)

        def addContainerMemory(memory):
            if memory != '': 
                command.append(MEMORY + '=' + memory)

        def addContainerCPUs(cpus):
            if cpus != '': 
                command.append(CPUS + '=' + cpus)

        def addContainerImage(image):
            command.append(image)

        def addRunCommand(runCommand):
            command.append(runCommand)

        def buildCommand():
            return " ".join(command)

        if not self.__imageExists(dockerImage):
            logging.info(f"Image {dockerImage} not found, pulling from remote repository...")
            self.__pullImage(dockerImage)
        
        if dockerCommand == '':
            addDockerRun()
            addRunOptions()
            addNetwork()
            addContainerName()
            addPrivileged()
            addDNS(dns)
            addContainerMemory(memory)
            addContainerCPUs(cpus)
            addContainerImage(dockerImage)
            addRunCommand(runCommand)
    
        try:    
            if dockerCommand != '':
                subprocess.run(dockerCommand, shell=True, capture_output=True)            
            else:
                subprocess.run(buildCommand(), shell=True, capture_output=True)
        except Exception as ex:
            logging.error(f"Error while criating the container {self.getNodeName()}: {str(ex)}")
            raise NodeInstantiationFailed(f"Error while criating the container {self.getNodeName()}: {str(ex)}")
        
        self.__enableNamespace(self.getNodeName())

    # Brief: Verifies if the image exists
    # Params:
    #   String image: Tag of the Docker image 
    # Return:
    #   True if the image exists locally
    def __imageExists(self, image: str) -> bool:
        out = subprocess.run(f"docker inspect --type=image {image}", shell=True, capture_output=True)
        outJson = json.loads(out.stdout.decode('utf-8'))
        if outJson == []: return False
        else: return True

            
    # Brief: Pulls the image from a Docker Hub repository
    # Params:
    #   String image: Tag of the Docker image 
    # Return:
    #   True if the image exists locally
    def __pullImage(self, image):
        try: 
            subprocess.run(f"docker pull {image}", shell=True)
        except Exception as ex:
            logging.error(f"Error pulling non-existing {image} image: {str(ex)}")
            raise NodeInstantiationFailed(f"Error pulling non-existing {image} image: {str(ex)}")

    # Brief: Instantiate the container
    # Params:
    #   String dockerImage: Name of the container of a local image or in a Docker Hub repository
    #   String DockerCommand: String to be used to instantiate the container instead of the standard command
    # Return:
    #   None
    def delete(self) -> None:
        try:    
            subprocess.run(f"docker kill {self.getNodeName()} && docker rm {self.getNodeName()}", shell=True, capture_output=True)
        except Exception as ex:
            logging.error(f"Error while deleting the host {self.getNodeName()}: {str(ex)}")
            raise NodeInstantiationFailed(f"Error while deleting the host {self.getNodeName()}: {str(ex)}")

    # Brief: Set Ip to an interface (the ip must be set only after connecting it to a container)
    # Params:
    #   String ip: IP address to be set to peerName interface
    #   int mask: Integer that represents the network mask
    #   String node: Reference to the node it is connected to this container to discover the intereface to set the ip to
    # Return:
    #   None
    def setIp(self, ip: str, mask: int, interfaceName: str) -> None:
        if not self.__interfaceExists(interfaceName):
            logging.error(f"Network interface {interfaceName} does not exist")
            raise Exception(f"Network interface {interfaceName} does not exist")

        self.__setIp(ip, mask, interfaceName)

    # Brief: Creates Linux virtual interfaces and connects peers to the nodes, in case of one of the nodes is a switch, it also creates a port in bridge
    # Params:
    #   Node node: Reference of another node to connect to
    # Return:
    #   None
    def connect(self, node: Node, interfaceName: str, peerInterfaceName: str) -> None:
        if self.__interfaceExists(interfaceName) or self.__interfaceExists(peerInterfaceName):
            logging.error(f"Cannot connect to {node.getNodeName()}, {interfaceName} or {peerInterfaceName} already exists")
            raise Exception(f"Cannot connect to {node.getNodeName()}, {interfaceName} or {peerInterfaceName} already exists")

        self.__create(interfaceName, peerInterfaceName)
        self.__setInterface(self.getNodeName(), interfaceName)
        self.__setInterface(node.getNodeName(), peerInterfaceName)

        if self.__class__.__name__ == 'Switch':
            self._Switch__createPort(self.getNodeName(), self.__getThisInterfaceName(node))
        if node.__class__.__name__ == 'Switch':
            node._Switch__createPort(node.getNodeName(), node.__getThisInterfaceName(self))
    
    def connectToInternet(self, hostIP: str, hostMask: int, interfaceName: str, hostInterfaceName: str) -> None:
        self.__create(interfaceName, hostInterfaceName)
        self.__setInterface(self.getNodeName(), interfaceName)
        if self.__class__.__name__ == 'Switch':
            self._Switch__createPort(self.getNodeName(), interfaceName)
        
        subprocess.run(f"ip link set {hostInterfaceName} up", shell=True)
        subprocess.run(f"ip addr add {hostIP}/{hostMask} dev {hostInterfaceName}", shell=True)

        # Enable forwading packets from host to interface
        hostGatewayInterfaceName = subprocess.run(f"route | grep \'^default\' | grep -o \'[^ ]*$\'", shell=True, capture_output=True).stdout.decode('utf8').replace("\n", '')
        subprocess.run(f"iptables -t nat -I POSTROUTING -o {hostGatewayInterfaceName} -j MASQUERADE", shell=True)
        subprocess.run(f"iptables -t nat -I POSTROUTING -o {hostInterfaceName} -j MASQUERADE", shell=True)
        subprocess.run(f"iptables -A FORWARD -i {hostInterfaceName} -o {hostGatewayInterfaceName} -j ACCEPT", shell=True)
        subprocess.run(f"iptables -A FORWARD -i {hostGatewayInterfaceName} -o {hostInterfaceName} -j ACCEPT", shell=True)
        subprocess.run(f"firewall-cmd --zone=trusted --add-interface={hostInterfaceName}", shell=True)
            
    def connectToInternetWithoutNAT(self, hostIP: str, hostMask: int, interfaceName: str, hostInterfaceName: str) -> None:
        self.__create(interfaceName, hostInterfaceName)
        self.__setInterface(self.getNodeName(), interfaceName)
        if self.__class__.__name__ == 'Switch':
            self._Switch__createPort(self.getNodeName(), interfaceName)
        
        subprocess.run(f"ip link set {hostInterfaceName} up", shell=True)
        subprocess.run(f"ip addr add {hostIP}/{hostMask} dev {hostInterfaceName}", shell=True)
        subprocess.run(f"firewall-cmd --zone=trusted --add-interface={hostInterfaceName}", shell=True)

    def enableForwarding(self, interfaceName: str, otherInterfaceName: str) -> None:
        self.run(f"iptables -t nat -I POSTROUTING -o {otherInterfaceName} -j MASQUERADE")
        self.run(f"iptables -t nat -I POSTROUTING -o {interfaceName} -j MASQUERADE")
        self.run(f"iptables -A FORWARD -i {interfaceName} -o {otherInterfaceName} -j ACCEPT")
        self.run(f"iptables -A FORWARD -i {otherInterfaceName} -o {interfaceName} -j ACCEPT")

    # Brief: Returns the value of the container name
    # Params:
    # Return:
    #   Returns the name of the node
    def getNodeName(self) -> str:
        return self.__nodeName

    # Brief: Add a route in routing table of container
    # Params:
    #   String ip: IP address of the route
    #   String mask: Network mask for the IP address route
    #   String interfaceName: Name of the interface to forward to
    # Return:
    #   None
    def addRoute(self, ip: str, mask: int,  interfaceName: str):
        if not self.__interfaceExists(interfaceName):
            logging.error(f"Network interface {interfaceName} does not exist")
            raise Exception(f"Network interface {interfaceName} does not exist")
        try:
            subprocess.run(f"docker exec {self.getNodeName()} ip route add {ip}/{mask} dev {interfaceName}", shell=True)
        except Exception as ex:
            logging.error(f"Error adding route {ip}/{mask} via {interfaceName} in {self.getNodeName()}: {str(ex)}")
            raise Exception(f"Error adding route {ip}/{mask} via {interfaceName} in {self.getNodeName()}: {str(ex)}")

    # Brief: Add a route in routing table of host
    # Params:
    #   String ip: IP address of the route
    #   String mask: Network mask for the IP address route
    #   String interfaceName: Name of the interface to forward to
    # Return:
    #   None
    def addRouteOnHost(self, ip: str, mask: int, interfaceName: str, gateway="0.0.0.0") -> None:
        try:
            subprocess.run(f"ip route add {ip}/{mask} via {gateway} dev {interfaceName}", shell=True)
        except Exception as ex:
            logging.error(f"Error adding route {ip}/{mask} via {interfaceName} in {self.getNodeName()}: {str(ex)}")
            raise Exception(f"Error adding route {ip}/{mask} via {interfaceName} in {self.getNodeName()}: {str(ex)}")


    # Brief: Set Ip to an interface (the ip must be set only after connecting it to a container, because)
    # Params:
    #   String destinationIp: The destination IP address of the gateway in format "XXX.XXX.XXX.XXX"
    #   String node: Reference to node that will serve as the first hop to forward the packets to the gateway, this reference node must be already connected to it
    # Return:
    #   None
    def setDefaultGateway(self, destinationIp: str, interfaceName: str) -> None:
        if not self.__interfaceExists(interfaceName):
            logging.error(f"Network interface {interfaceName} does not exist")
            raise Exception(f"Network interface {interfaceName} does not exist")
        
        self.addRoute(destinationIp, 32, interfaceName)
        try:
            subprocess.run(f"docker exec {self.getNodeName()} route add default gw {destinationIp} dev {interfaceName}", shell=True)
        except Exception as ex:
            logging.error(f"Error while setting gateway {destinationIp} on device {interfaceName} in {self.getNodeName()}: {str(ex)}")
            raise Exception(f"Error while setting gateway {destinationIp} on device {interfaceName} in {self.getNodeName()}: {str(ex)}")

    # Brief: Runs a command inside the container
    # Params:
    #   String command: String containing the command to run inside the container
    # Return:
    #   Returns variable that contains stdout and stderr (more information in subprocess documentation)
    def run(self, command: str) -> str:
        try:
            command = command.replace('\"', 'DOUBLEQUOTESDELIMITER')
            command = f"docker exec {self.getNodeName()} bash -c \"" + command + f"\""
            command = command.replace('DOUBLEQUOTESDELIMITER','\\"')
            return subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, text=True)
        except Exception as ex:
            logging.error(f"Error executing command {command} in {self.getNodeName()}: {str(ex)}")
            raise Exception(f"Error executing command {command} in {self.getNodeName()}: {str(ex)}")

    # Brief: Runs multiple commands inside the container
    # Params:
    #   List<String> commands: Runs multiple comands inside the container
    # Return:
    #   Returns a list with the variable that contains stdout and stderr (more information in subprocess documentation)
    def runs(self, commands: list) -> list:
        return [self.run(command) for command in commands]     

    # Brief: Copy local file into container
    # Params:
    #   String path: Absolute or relative path to the file to be copied from local (path+filename)
    #   String destPath: Absolute path to copy the file to the container (path+filename)
    # Return:
    def copyLocalToContainer(self, path: str, destPath: str) -> None:
        try:
            subprocess.run(f"docker cp {path} {self.getNodeName()}:{destPath}", shell=True, capture_output=True)
        except Exception as ex:
            logging.error(f"Error copying file from {path} to {destPath}: {str(ex)}")
            raise Exception(f"Error copying file from {path} to {destPath}: {str(ex)}")

    # Brief: Copy local file into container
    # Params:
    #   String path: Absolute path to the file to be copied from container (path+filename)
    #   String destPath: Absolute or relative path to copy to local (path+filename)
    # Return:
    def copyContainerToLocal(self, path: str, destPath: str) -> None:
        try:
            subprocess.run(f"docker cp {self.getNodeName()}:{path} {destPath}", shell=True, capture_output=True)
        except Exception as ex:
            logging.error(f"Error copying file from {path} to {destPath}: {str(ex)}")
            raise Exception(f"Error copying file from {path} to {destPath}: {str(ex)}")

    def __interfaceExists(self, interfaceName: str) -> bool:
        out = subprocess.run(f"docker exec {self.getNodeName()} ip link | grep {interfaceName}", shell=True, capture_output=True)
        if out.stdout.decode("utf8") != '':
            return True
        return False

    # Brief: Returns the name of the interface to be created on this node
    # Params:
    #   Node node: Reference of another node to connect to
    # Return:
    #   Name of the interface with pattern veth + this node name + other node name
    def __getThisInterfaceName(self, node: Node) -> str:
        return self.getNodeName()+node.getNodeName()

    # Brief: Set Ip to an interface
    # Params:
    #   String ip: IP address to be set to peerName interface
    #   int mask: Integer that represents the network mask
    #   String interfaceName: Name of the interface to set the ip
    # Return:
    #   None
    def __setIp(self, ip: str, mask: int, interfaceName: str) -> None:
        try:
            subprocess.run(f"ip -n {self.getNodeName()} addr add {ip}/{mask} dev {interfaceName}", shell=True)
        except Exception as ex:
            logging.error(f"Error while setting IP {ip}/{mask} to virtual interface {interfaceName}: {str(ex)}")
            raise Exception(f"Error while setting IP {ip}/{mask} to virtual interface {interfaceName}: {str(ex)}")

    # Brief: Returns the name of the interface to be created on other node
    # Params:
    #   Node node: Reference of another node to connect to
    # Return:
    #   Name of the interface with pattern veth + other node name + this node name
    def __getOtherInterfaceName(self, node: Node) -> str:
        return node.getNodeName()+self.getNodeName()

    # Brief: Creates the virtual interfaces and set them up (names cant be the same as some existing one in host's namespace)
    # Params:
    #   String peer1Name: Name of the interface to connect to the first peer 
    #   String peer2Name: Name of the interface to connect to the second peer 
    # Return:
    #   None
    def __create(self, peer1Name: str, peer2Name: str) -> None:
        try:
            subprocess.run(f"ip link add {peer1Name} type veth peer name {peer2Name}", shell=True)
        except Exception as ex:
            logging.error(f"Error while creating virtual interfaces {peer1Name} and {peer2Name}: {str(ex)}")
            raise Exception(f"Error while creating virtual interfaces {peer1Name} and {peer2Name}: {str(ex)}")

    # Brief: Set the interface to node
    # Params:
    #   String nodeName: Name of the node network namespace
    #   String peerName: Name of the interface to set to node
    # Return:
    #   None
    def __setInterface(self, nodeName: str, peerName: str) -> None:
        try:
            subprocess.run(f"ip link set {peerName} netns {nodeName}", shell=True)
            subprocess.run(f"ip -n {nodeName} link set {peerName} up", shell=True)
        except Exception as ex:
            logging.error(f"Error while setting virtual interfaces {peerName} to {nodeName}: {str(ex)}")
            raise Exception(f"Error while setting virtual interfaces {peerName} to {nodeName}: {str(ex)}")

    # Brief: Enable accessing the Docker node namespace directly
    # Params:
    # Return:
    #   None
    def __enableNamespace(self, nodeName) -> None:
        try:    
            subprocess.run(f"pid=$(docker inspect -f '{{{{.State.Pid}}}}' {nodeName}); mkdir -p /var/run/netns/; ln -sfT /proc/$pid/ns/net /var/run/netns/{nodeName}", shell=True)
        except Exception as ex:
            logging.error(f"Error while deleting the host {self.getNodeName()}: {str(ex)}")
            raise Exception(f"Error while deleting the host {self.getNodeName()}: {str(ex)}")

    # Brief: Get all interfaces names
    # Params:
    # Return:
    #   Return a list with the name of all interfaces
    def __getAllIntefaces(self) -> list:
        output = subprocess.run(f"docker exec {self.getNodeName()} ifconfig -a | sed 's/[ \t].*//;/^$/d'", shell=True, capture_output=True)
        interfaces=output.stdout.decode('utf8').replace(":", '').split('\n')
        return list(filter(None, interfaces)) # Remove empty strings

    # Brief: Verifies if the container is active
    # Params:
    # Return:
    #   Return true if it is active or false otherwise
    def __isActive(self) -> bool:
        if subprocess.run(f"docker ps | grep {self.getNodeName()}'", shell=True, capture_output=True).stdout.decode('utf8') != '': return True
        return False

    def setMtuSize(self, interfaceName: str, mtu: int) -> None:
        self.run(f"ifconfig {interfaceName} mtu {str(mtu)}")

    # Brief: It retrieves the config file inside the container and reads it, and then returns a ConfigParser instance
    # Params:
    #   String containerPath: Path inside the container of the config file including filename
    # Return:
    #   Returns a ConfigParser instance with the config file read
    def readConfigFile(self, containerPath: str) -> None:
        randomTmpName = self.getHashFromString(containerPath) 
        self.copyContainerToLocal(containerPath, f"/tmp/lft/{randomTmpName}")
        
        config = ConfigParser()
        config.read(f"/tmp/lft/{randomTmpName}")
        return config

    def saveConfig(self, config: ConfigParser, containerPath: str) -> None:
        randomTmpName = self.getHashFromString(containerPath)
        with open(f"/tmp/lft/{randomTmpName}", "w") as f:
            config.write(f)
        self.copyLocalToContainer(f"/tmp/lft/{randomTmpName}", containerPath)
        
    def getHashFromString(self, anyStr: str) -> str:
        h = hashlib.md5()
        h.update(anyStr.encode('utf-8'))
        return h.hexdigest()
    
    def setHost(self, ip: str) -> None:
        result = self.run(f"hostname")
        hostname = result.stdout.read().replace("\n", "")
        self.run(f"HOSTNAME=$(hostname) && echo \"{ip} {hostname}\" >> /etc/hosts")

    def acceptPacketsFromInterface(self, interfaceName: str):
        self.run(f"iptables -A INPUT -i {interfaceName} -j ACCEPT")

    # Brief: Shapes the Linux network interface throughput, latency and burst using Traffic Control (TC)
    # Params:
    #   String interfaceName: Name of the interface to set the traffic control
    #   int throughput: Value of the throughput in bits
    #    
    # Return:
    #   Returns a ConfigParser instance with the config file read
    def setInterfaceProperties(self, interfaceName: str, throughput: str, delay: str, jitter: str) -> None:
        self.run(f"tc qdisc add dev {interfaceName} root netem delay {delay} {jitter} rate {throughput}")


from profissa_lft.switch import Switch
from profissa_lft.ue import UE
from profissa_lft.epc import EPC
from profissa_lft.enb import EnB
import time

epc = EPC('epc')
enb = EnB("enb")
ue2 = UE('ue2')
ue1 = UE('ue1')

epc.instantiate(dockerImage="alexandremitsurukaihara/lft:perfsonar-toolkit-srsRAN")
enb.instantiate(dockerImage="alexandremitsurukaihara/lft:perfsonar-toolkit-srsRAN")
ue2.instantiate(dockerImage="alexandremitsurukaihara/lft:perfsonar-toolkit-srsRAN")
ue1.instantiate(dockerImage="alexandremitsurukaihara/lft:perfsonar-toolkit-srsRAN")

enb.connect(epc, "enbepc", "epcenb")
ue1.connect(enb, "ue1enb", "enbue1")
ue2.connect(enb, "ue2enb", "enbue2")

epc.setIp('10.0.0.1', 24, "epcenb")
enb.setIp('10.0.0.2', 24, "enbepc")
enb.setIp('11.0.0.1', 30, "enbue1")
enb.setIp('11.0.0.5', 30, "enbue2")
ue1.setIp('11.0.0.2', 29, "ue1enb")
ue2.setIp('11.0.0.6', 29, "ue2enb")

enb.connectToInternet('10.0.0.6', 24, "epchost", "hostepc")

epc.setDefaultGateway('10.0.0.6', "epchost")
enb.setDefaultGateway('10.0.0.6', "enbepc")
ue1.setDefaultGateway('10.0.0.2', "ue1enb")
ue2.setDefaultGateway('10.0.0.2', "ue2enb")

# Define EPC config
epc.setEPCAddress("10.0.0.1")
epc.addNewUE(ue1.getNodeName(), "001010123456780", "172.16.0.2")
epc.addNewUE(ue2.getNodeName(), "001010123456789", "172.16.0.3")

# Define ENB Config
enb.setEPCAddress("10.0.0.1")
enb.setEnBAddress("10.0.0.2")
enb.setMultiUEEnBAddr("11.0.0.1", 2101, '11.0.0.1', 2100)
enb.setMultiUEUE1Addr("11.0.0.2", 2001, "11.0.0.1", 2000)
enb.setMultiUEUE2Addr("11.0.0.6", 2011, "11.0.0.5", 2010)

# Define UE Config
ue1.setUEID("001010123456780")
ue2.setUEID("001010123456789")

epc.start()
enb.starGnuRadioMultiUE()
enb.start("11.0.0.1", 2101, "11.0.0.1", 2100)
time.sleep(5)  
ue1.start("11.0.0.2", 2001, "11.0.0.1", 2000)
ue2.start("11.0.0.6", 2011, "11.0.0.5", 2010)



"""
Mininet Cleanup
author: Bob Lantz (rlantz@cs.stanford.edu)

Unfortunately, Mininet and OpenFlow (and the Linux kernel)
don't always clean up properly after themselves. Until they do
(or until cleanup functionality is integrated into the Python
code), this script may be used to get rid of unwanted garbage.
It may also get rid of 'false positives', but hopefully
nothing irreplaceable!
"""

from __future__ import unicode_literals
try:
	from builtins import str
except ImportError:
	pass
from subprocess import Popen
from subprocess import call
import time
import os
import iptc
import ipaddress
import shlex

from mininet.log import info
from mininet.term import cleanUpScreens
from mininet.clean import sh, killprocs
from containernet.net import SAP_PREFIX
from mn_wifi.clean import Cleanup as Cleanup_nm_wifi


class Cleanup( object ):
    "Wrapper for cleanup()"

    callbacks = []

    @classmethod
    def cleanup( cls):
        """Clean up junk which might be left over from old runs;
           do fast stuff before slow dp and link removal!"""

        Cleanup_nm_wifi.cleanup_wifi()

        info( "**** Removing excess controllers/ofprotocols/ofdatapaths/"
              "pings/noxes\n" )
        zombies = ( 'controller ofprotocol ofdatapath ping nox_core'
                    'lt-nox_core ovs-openflowd ovs-controller'
                    'ovs-testcontroller udpbwtest mnexec ivs ryu-manager' )
        # Note: real zombie processes can't actually be killed, since they
        # are already (un)dead. Then again,
        # you can't connect to them either, so they're mostly harmless.
        # Send SIGTERM first to give processes a chance to shutdown cleanly.
        sh( 'killall ' + zombies + ' 2> /dev/null' )
        time.sleep( 1 )
        sh( 'killall -9 ' + zombies + ' 2> /dev/null' )

        # And kill off sudo mnexec
        sh( 'pkill -9 -f "sudo mnexec"')

        info( "*** Removing junk from /tmp\n" )
        sh( 'rm -f /tmp/vconn* /tmp/vlogs* /tmp/*.out /tmp/*.log' )

        info( "*** Removing old X11 tunnels\n" )
        cleanUpScreens()

        info( "*** Removing excess kernel datapaths\n" )
        dps = sh( "ps ax | egrep -o 'dp[0-9]+' | sed 's/dp/nl:/'"
                  ).splitlines()
        for dp in dps:
            if dp:
                sh( 'dpctl deldp ' + dp )

        info( "***  Removing OVS datapaths\n" )
        dps = sh("ovs-vsctl --timeout=1 list-br").strip().splitlines()
        if dps:
            sh( "ovs-vsctl " + " -- ".join( "--if-exists del-br " + dp
                                            for dp in dps if dp ) )
        # And in case the above didn't work...
        dps = sh( "ovs-vsctl --timeout=1 list-br" ).strip().splitlines()
        for dp in dps:
            sh( 'ovs-vsctl del-br ' + dp )

        info( "*** Removing all links of the pattern foo-ethX\n" )
        links = sh( "ip link show | "
                    "egrep -o '([-_.[:alnum:]]+-eth[[:digit:]]+)'"
                    ).splitlines()
        # Delete blocks of links
        n = 1000  # chunk size
        for i in range( 0, len( links ), n ):
            cmd = ';'.join( 'ip link del %s' % link
                             for link in links[ i : i + n ] )
            sh( '( %s ) 2> /dev/null' % cmd )

        if 'tap9' in sh( 'ip link show' ):
            info( "*** Removing tap9 - assuming it's from cluster edition\n" )
            sh( 'ip link del tap9' )

        info( "*** Killing stale mininet node processes\n" )
        killprocs( 'mininet:' )

        info( "*** Shutting down stale tunnels\n" )
        killprocs( 'Tunnel=Ethernet' )
        killprocs( '.ssh/mn')
        sh( 'rm -f ~/.ssh/mn/*' )

        # Call any additional cleanup code if necessary
        for callback in cls.callbacks:
            callback()

        # Containernet should also cleanup pending Docker
        cmd =  "docker rm -f $( docker ps --filter 'label=com.containernet' -a -q)"
        call(cmd, shell=True, stdout=open(os.devnull, 'wb'),  stderr=open(os.devnull, 'wb'))

        # cleanup any remaining iptables rules from external SAPs with NAT
        # we use iptc module to iterate through the loops, but due to a bug, we cannot use iptc to delete the rules
        # we rely on iptables CLI to delete the found rules
        info("***  Removing SAP NAT rules\n")
        table = iptc.Table(iptc.Table.NAT)
        chain = iptc.Chain(table, 'POSTROUTING')

        for rule in chain.rules:
            if SAP_PREFIX in str(rule.out_interface):
                src_CIDR = str(ipaddress.IPv4Network(u'{}'.format(str(rule.src))))
                rule0_ = "iptables -t nat -D POSTROUTING ! -o {0} -s {1} -j MASQUERADE".\
                    format(rule.out_interface.strip('!'), src_CIDR)
                p = Popen(shlex.split(rule0_))
                p.communicate()
                info("delete NAT rule from SAP: {1} - {0} - {2}\n".format(rule.out_interface, rule.in_interface, src_CIDR))


        table = iptc.Table(iptc.Table.FILTER)
        chain = iptc.Chain(table, 'FORWARD')
        for rule in chain.rules:
            src_CIDR = str(ipaddress.IPv4Network(u'{}'.format(str(rule.src))))
            if SAP_PREFIX in str(rule.out_interface):
                rule1_ = "iptables -D FORWARD -o {0} -j ACCEPT".format(rule.out_interface)
                p = Popen(shlex.split(rule1_))
                p.communicate()
                info("delete FORWARD rule from SAP: {1} - {0} - {2}\n".format(rule.out_interface, rule.in_interface,
                                                                              src_CIDR))

            if SAP_PREFIX in str(rule.in_interface):
                rule2_ = "iptables -D FORWARD -i {0} -j ACCEPT".format(rule.in_interface)
                p = Popen(shlex.split(rule2_))
                p.communicate()
                info("delete FORWARD rule from SAP: {1} - {0} - {2}\n".format(rule.out_interface, rule.in_interface,
                                                                              src_CIDR))

        info( "*** Cleanup complete.\n" )

    @classmethod
    def addCleanupCallback( cls, callback ):
        "Add cleanup callback"
        if callback not in cls.callbacks:
            cls.callbacks.append( callback )


cleanup = Cleanup.cleanup
addCleanupCallback = Cleanup.addCleanupCallback



from subprocess import run

class PSchedulerWrapper:
    def __init__(self, centralArchiverPath = "central_archiver.json"):
        self.program = "pscheduler"
        self.programOption = ""
        self.command = ""
        self.archiverOption = "--archiver"
        self.maxRunsOption = "--max-runs"
        self.repeatOption = "--repeat"
        self.outputFileOption = "--output"
        self.formatOption = "--format"
        self.options = []
        self.archiverPath = centralArchiverPath
        self.archiverConf = ""


    def MaxRuns(self, maxRuns: int) -> None:
        option = self.joinClauses(self.maxRunsOption, str(maxRuns), sep="=")
        self.addOption(option)
        return self
    
    def Repeat(self, interval: str) -> None:
        option = self.joinClauses(self.repeatOption, interval, sep="=")
        self.addOption(option)
        return self

    def Archiver(self) -> None:
        option = self.joinClauses(self.archiverOption, self.archiverConf)
        self.addOption(option)
        return self

    def OutputFile(self, path="", fileName=""):
        option = self.joinClauses(self.outputFileOption, path + fileName, sep="=")
        self.addOption(option)
        return self
    
    def Format(self, format):
        option = self.joinClauses(self.formatOption, format)
        self.addOption(option)
        return self

    def Command(self, command: str) -> None:
        self.command = command

    def ProgramOption(self, option):
        self.programOption = option

    def joinClauses(self, *args, sep = " ") -> str:
        if(isinstance(args[0], list)):
            return sep.join(args[0])
        return sep.join(args)

    def addOption(self, option) -> None:
        self.options.append(option)

    def __loadArchiverConf(self) -> None:
        with open(self.archiverPath, "r") as f:
            self.archiverConf = f.read()

    def run(self):
        run(self.command, shell=True)


class Task(PSchedulerWrapper):
    def __init__(self, centralArchiverPath = "central_archiver.json"):
        super().__init__(centralArchiverPath)
        self.ProgramOption("task")
        self.taskType = ""
        self.taskOptions = []
        self.sourceOption = "--source"
        self.destinationOption = "--dest"
        self.durationOption = "-t"
        self.intervalOption = "-i"
        self.LATENCY = "latency"
        self.RTT = "rtt"
        self.THROUGHPUT = "throughput"

    def TaskType(self, taskName):
        taskTypeNames = [self.LATENCY, self.RTT, self.THROUGHPUT]
        if taskName not in taskTypeNames:
            raise ValueError(f"{taskName} is not a valid name, choose one of {taskTypeNames}.")
        self.taskType = taskName
        return self

    def Source(self, sourceIp: str) -> None:
        option = self.joinClauses(self.sourceOption, sourceIp)
        self.addTaskOption(option)
        return self

    def Dest(self, destIp: str) -> None:
        option = self.joinClauses(self.destinationOption, destIp)
        self.addTaskOption(option)
        return self
    
    def Duration(self, duration):
        option = self.joinClauses(self.durationOption, duration)
        self.addOption(option)
        return self
    
    def Interval(self, interval):
        option = self.joinClauses(self.intervalOption, interval)
        self.addTaskOption(option)
        return self
    
    def addTaskOption(self, option):
        self.taskOptions.append(option)

    def mountCommand(self) -> str:
        options = self.joinClauses(self.options)
        taskOptions = self.joinClauses(self.taskOptions)
        self.Command(self.joinClauses(self.program, self.programOption, options, self.taskType, taskOptions))
        return self

    def getCommand(self):
        return self.command

class Throughput(Task):
    def __init__(self, centralArchiverPath = "central_archiver.json"):
        super().__init__(centralArchiverPath)
        self.TaskType(self.THROUGHPUT)
        self.durationOption = "--duration"

    def ThroughputDuration(self, duration: int) -> None:
        option = self.joinClauses(self.durationOption, str(duration))
        self.addTaskOption(option)
        return self

class Latency(Task):
    def __init__(self, centralArchiverPath = "central_archiver.json"):
        super().__init__(centralArchiverPath)
        self.TaskType(self.LATENCY)
        self.packetCountOption = "-c"
        self.packetIntervalOption = "-i"
        self.OUTPUT_RAW = "-R"
        
    def PacketCount(self, packetCount):
        option = self.joinClauses(self.packetCountOption, str(packetCount))
        self.addTaskOption(option)
        return self

    def PacketInterval(self, packetInterval):
        option = self.joinClauses(self.packetIntervalOption, packetInterval)
        self.addTaskOption(option)
        return self
    
    def OutputRaw(self):
        self.addTaskOption(self.OUTPUT_RAW)
        return self

class Rtt(Task):
    def __init__(self, centralArchiverPath = "central_archiver.json"):
        super().__init__(centralArchiverPath)
        self.TaskType(self.RTT)
        self.countOption = "--count"

    def Count(self, count: int) -> None:
        option = self.joinClauses(self.countOption, str(count))
        self.addTaskOption(option)
        return self


from profissa_lft.host import Host
from profissa_lft.switch import Switch
from profissa_lft.controller import Controller

h1 = Host('h1')
h2 = Host('h2')
s1 = Switch('s1')
c1 = Controller('c1')

h1.instantiate()
h2.instantiate()
s1.instantiate()
c1.instantiate()

h1.connect(s1, "h1s1", "s1h1")
h2.connect(s1, "h2s1", "s1h2")
c1.connect(s1, "c1s1", "s1c1")

h1.setIp('10.0.0.1', 24, "h1s1")
h2.setIp('10.0.0.2', 24, "h2s1")
c1.setIp('10.0.0.3', 24, "c1s1")

c1.initController('10.0.0.3', 9001)
s1.setController('10.0.0.3', 9001)

s1.connectToInternet('10.0.0.4', 24, "s1host", "hosts1")

h1.setDefaultGateway('10.0.0.4', "h1s1")
h2.setDefaultGateway('10.0.0.4', "h2s1")
c1.setDefaultGateway('10.0.0.4', "c1s1")



# Copyright (C) 2023 Alexandre Mitsuru Kaihara
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .perfsonar import Perfsonar
from configparser import ConfigParser
from .constants import *
import pandas as pd

class EPC(Perfsonar):
    def __init__(self, name: str):
        super().__init__(name)
        self.defaultEPCConfigPath = '/etc/srsran/epc.conf'
        self.configEPC = None
        self.defaultEPCUserDbPath = '/etc/srsran/user_db.csv'
        self.userDb = None 
        self.buildDir = "/srsRAN/build"

    def instantiate(self, dockerImage='alexandremitsurukaihara/lft:srsran', runCommand='') -> None:
        super().instantiate(dockerImage=dockerImage, runCommand=runCommand)
        self.configEPC = self.readConfigFile(self.defaultEPCConfigPath)
        self.userDb = self.createUserDb()

    def start(self) -> None:
        self.run(f"{self.buildDir}/srsepc/src/srsepc > {self.buildDir}/epc.out &")

    def stop(self) -> None:
        self.run("pkill -f -9 srsepc")

    def setDefaultEPCConfigPath(self, path: str) -> None:
        self.defaultEPCConfigPath = path

    def getDefaultEPCConfigPath(self) -> str:
        return self.defaultEPCConfigPath

    def setEPCAddress(self, ip='127.0.1.100') -> None:
        self.configEPC[MME_SECTION][MME_BIND_ADDR] = ip
        self.configEPC[SPGW_SECTION][GTPU_BIND_ADDR] = ip
        self.saveConfig(self.configEPC, self.defaultEPCConfigPath)
        
    def setSgiInterfaceAddress(self, ip='172.16.0.1') -> None:
        self.configEPC[SPGW_SECTION][SGI_IF_ADDR] = ip
        self.saveConfig(self.configEPC, self.defaultEPCConfigPath)
        
    # Each UE ID must be unique and must be set in "imsi" parameter located inside the ue.conf
    def addNewUE(self, name: str, ID: str, IP="dynamic") -> None:
        newRow = {
            "Name": name,
            "Auth": "mil",
            "IMSI": ID,
            "Key": "00112233445566778899aabbccddeeff",
            "OP_Type": "opc",
            "OP/OPc": "63bfa50ee6523365ff14c1f45f88737d",
            "AMF": "9001",
            "SQN": "000000001234",
            "QCI": "7",
            "IP_alloc": IP
        }
        self.userDb.loc[len(self.userDb)] = newRow
        self.saveUserDb()
        #self.run(f"echo \'{name},mil,{ID},00112233445566778899aabbccddeeff,opc,63bfa50ee6523365ff14c1f45f88737d,9001,000000001234,7,{IP}\' >> {self.defaultEPCUserDbPath}")

    def createUserDb(self) -> None:
        return pd.DataFrame(columns=["Name","Auth","IMSI","Key","OP_Type","OP/OPc","AMF","SQN","QCI","IP_alloc"]) 
    
    def saveUserDb(self) -> None:
        randomTmpName = self.getHashFromString(self.defaultEPCUserDbPath)
        self.userDb.to_csv(f"/tmp/lft/{randomTmpName}", header=None, index=False)
        self.copyLocalToContainer(f"/tmp/lft/{randomTmpName}", self.defaultEPCUserDbPath)



from .experiment import *
from .emu_emu_wired import EmuEmuWired
from .emu_phy_wired import EmuPhyWired


# wired emu-emu
emuEmu = EmuEmuWired()
emuEmu.setup()
runThroughput(EMU_EMU_WIRED_PREFIX, EMU_EMU_WIRED_H1_IP, EMU_EMU_WIRED_H2_IP)
runRTT(EMU_EMU_WIRED_PREFIX, EMU_EMU_WIRED_H1_IP, EMU_EMU_WIRED_H2_IP)
runLatency(EMU_EMU_WIRED_PREFIX, EMU_EMU_WIRED_H1_IP, EMU_EMU_WIRED_H2_IP)
emuEmu.tearDown()

# wired phy-emu
emuPhy = EmuPhyWired()
emuPhy.setup()
runThroughput(EMU_PHY_WIRED_PREFIX, EMU_PHY_WIRED_H1_IP, EMU_PHY_WIRED_H2_IP)
runRTT(EMU_PHY_WIRED_PREFIX, EMU_PHY_WIRED_H1_IP, EMU_PHY_WIRED_H2_IP)
runLatency(EMU_PHY_WIRED_PREFIX, EMU_PHY_WIRED_H1_IP, EMU_PHY_WIRED_H2_IP)
emuPhy.tearDown()

# wired phy-phy
runThroughput(PHY_PHY_WIRED_PREFIX, PHY_PHY_WIRED_H1_IP, PHY_PHY_WIRED_H2_IP)
runRTT(PHY_PHY_WIRED_PREFIX, PHY_PHY_WIRED_H1_IP, PHY_PHY_WIRED_H2_IP)
runLatency(PHY_PHY_WIRED_PREFIX, PHY_PHY_WIRED_H1_IP, PHY_PHY_WIRED_H2_IP)


from profissa_lft.host import Host
from profissa_lft.switch import Switch


class DeployLFT():
	def deploy(self, size):
		s1 = Switch("s1")
		s1.instantiate()
		[self.__addHost(i, s1) for i in range(size)]

	def __addHost(self, counter, switch):
		host = Host(f"h{counter}")
		host.instantiate(dockerImage="ubuntu:trusty", runCommand="tail -f /dev/null")
		switch.connect(host, f"s1h{counter}", f"h{counter}s1")
		host.setIp(f"10.0.{int(counter/256)}.{(counter+2)%256}", 24, f"h{counter}s1")

	def getReferences(self, size):
		self.nodes = []
		self.nodes.append(Switch("s1"))
		[self.nodes.append(Host(f"h{i}")) for i in range(size)]

	def undeploy(self):
		[node.delete() for node in self.nodes]



from containernet.link import TCLink
from containernet.net import Containernet


class DeployMininet():
        def deploy(self, size):
                self.net = Containernet()
                s1 = self.net.addSwitch("s1")
                [self.__addHost(i, s1) for i in range(size)]
                self.net.start()
        def __addHost(self, counter, switch):
                image = "ubuntu:trusty"
                host = self.net.addDocker(f"d{counter+1}", ip=f"10.0.{int(counter/256)}.{(counter+2)%256}", dimage=image)
                self.net.addLink(host, switch)
        def undeploy(self):
                self.net.stop()




#!/usr/bin/env python
import csv
from email.mime import image
from gc import collect
import signal
import sys
import subprocess
from wsgiref.simple_server import ServerHandler, server_version
from host import Host
from switch import Switch
from glob import glob
from controller import Controller
from configparser import ConfigParser
from cicflowmeter import CICFlowMeter
from node import Node
from os import getcwd
from globalvariables import *

class Seafile(Host):
    def instantiate(self):
        super().instantiate(dockerImage=seafileserver)
    def updateServerConfig(self) -> None:
        self.copyContainerToLocal("/home/seafolder", "seafolder")
        out = subprocess.run("cat seafolder", shell=True, capture_output=True).stdout.decode('utf8')
        parser = ConfigParser()
        parser.read('serverconfig.ini')
        parser.set("50", "seafolder",  out)
        parser.set("200", "seafolder", out)
        parser.set("210", "seafolder", out)
        parser.set("220", "seafolder", out)
        with open('serverconfig.ini', 'w') as configfile:
            parser.write(configfile)


class LinuxClient(Host):
    def setAutomationScripts(self, path) -> None:
        self.copyLocalToContainer(path, "/home/debian/automation")
    def setPrinterIp(self, path) -> None:
        self.copyLocalToContainer(path, "/home/debian/automation/packages/system/printerip")
    def setSshIpList(self, path) -> None:
        self.copyLocalToContainer(path, "/home/debian/automation/packages/system/sshiplist.ini")
    def setClientBehaviour(self, path) -> None:
        self.copyLocalToContainer(path, "/home/debian/automation/packages/system/config.ini")
    def setServerConfig(self, path) -> None:
        self.copyLocalToContainer(path, "/home/debian/automation/packages/system/serverconfig.ini")
    def setIpListPort80(self, path) -> None:
        self.copyLocalToContainer(path, "/home/debian/automation/packages/attacking/ipListPort80.txt")
    def setIpList(self, path) -> None:
        self.copyLocalToContainer(path, "/home/debian/automation/packages/attacking/ipList.txt")
    def setIpRange(self, path) -> None:
        self.copyLocalToContainer(path, "/home/debian/automation/packages/attacking/iprange.txt")


def setLinuxClientFileConfig(node: LinuxClient, subnet: str, behaviour: str):
    printTerm(f"[LST2.0] Copying Configuration Files to Container {node.getNodeName()}")
    if subnet != external_subnet: aux = "internal"
    else: aux = "external"
    node.setAutomationScripts("automation")
    node.setPrinterIp(f"printersip/{subnet.split('.')[2]}")
    node.setSshIpList("sshiplist.ini")
    node.setClientBehaviour(f"client_behaviour/{behaviour}.ini")
    node.setServerConfig("serverconfig.ini")
    node.setIpListPort80(f"{aux}_ipListPort80.txt")        
    node.setIpList(f"{aux}_ipList.txt")
    node.setIpRange(f"{aux}_iprange.txt")


def setNetworkConfig(node: Node, bridge: Node, subnet: str, address: int, setFiles=True) -> Node:
    node.connect(bridge)
    printTerm(f"[LST2.0] ... Connecting to {bridge.getNodeName()}")
    node.setIp(subnet+str(address), 24, bridge)
    printTerm(f"[LST2.0] ... Setting IP {subnet+str(address)}")
    # Define default gateway of nodes
    if bridge == nodes['brint']: node.setDefaultGateway(int_gateway, bridge)
    if bridge == nodes['brex']:  node.setDefaultGateway(ex_gateway , bridge)
    printTerm(f"[LST2.0] ... Setting Default Gateway")
    # Add routes to enable nodes within internal subnet communicate with server subnet
    if subnet != server_subnet: node.addRoute(server_subnet+'0', 24, bridge)
    if subnet != management_subnet: node.addRoute(management_subnet+'0', 24, bridge)
    if subnet != office_subnet: node.addRoute(office_subnet+'0', 24, bridge)
    if subnet != developer_subnet: node.addRoute(developer_subnet+'0', 24, bridge)
    if subnet != external_subnet: node.addRoute(external_subnet+'0', 24, bridge)
    printTerm(f"[LST2.0] ... Adding new routes")
    if setFiles:
        subprocess.run(f"docker cp serverconfig.ini {node.getNodeName()}:/home/debian/serverconfig.ini", shell=True)
        subprocess.run(f"docker cp backup.py {node.getNodeName()}:/home/debian/backup.py", shell=True)


def createBridge(name: str, ip: str, gatewayIp: str, netflowPort=9000) -> None: 
    printTerm(f"[LST2.0] Creating Switch {name}")
    nodes[name] = Switch(name, getcwd()+'/flows/'+name, '/home/pcap')
    nodes[name].run('mkdir /home/pcap > /dev/null 2>&1')
    nodes[name].instantiate()
    printTerm("[LST2.0] ... Instantiating container")
    nodes[name].setIp(ip, 24)
    printTerm(f"[LST2.0] ... Setting IP {ip}")
    nodes[name].connectToInternet(gatewayIp, 24)
    printTerm(f"[LST2.0] ... Connecting to Internet through Host")


def createController(name: str, bridgeName: str, controllerIp: str, controllerPort: int) -> None:
    printTerm(f"[LST2.0] Creating Controller {name}")
    nodes[name] = Controller(name)
    nodes[name].instantiate()
    printTerm("[LST2.0] ... Instantiating container")
    nodes[name].connect(nodes[bridgeName])
    printTerm(f"[LST2.0] ... Connecting to Node {bridgeName}")
    nodes[name].setIp(controllerIp, 24, nodes[bridgeName])
    printTerm(f"[LST2.0] ... Setting IP {controllerIp}")
    nodes[name].initController(controllerIp, controllerPort)
    printTerm(f"[LST2.0] ... Initiating Ryu Controller at {controllerIp}:{controllerPort}")
    nodes[bridgeName].setController(controllerIp, controllerPort)
    printTerm(f"[LST2.0] ... Conecting Switch {bridgeName} to controller")
    

def createServer(name: str, serverImage: str, subnet: str,  address: int) -> None:
    printTerm(f"[LST2.0] Creating Server {name}")
    nodes[name] = Host(name)
    printTerm("[LST2.0] ... Instantiating container")
    nodes[name].instantiate(serverImage)
    setNetworkConfig(nodes[name], nodes['brint'], subnet, address)


def createLinuxClient(name, bridge: Node, subnet: str, address: int) -> None:
    printTerm(f"[LST2.0] Creating Client {name}")
    nodes[name] = LinuxClient(name)
    nodes[name].instantiate(linuxclient)
    printTerm("[LST2.0] ... Instantiating container")
    setNetworkConfig(nodes[name], bridge, subnet, address)


def createPrinter(name: str, subnet: str) -> None:
    printTerm(f"[LST2.0] Creating Printer {name}")
    nodes[name] = Host(name)
    nodes[name].instantiate(printerserver)
    printTerm("[LST2.0] ... Instantiating container")
    setNetworkConfig(nodes[name], nodes['brint'], subnet, 1)


def unmakeChanges():
    printTerm(f"[LST2.0] Unmaking Experiment. Deleting Containers")
    [node.delete() for _,node in nodes.items()]


def convertPcap():
    printTerm(f"[LST2.0] Converting pcap files with CICFlowMeter")
    pcaps = glob('flows/brint/*')
    pcaps = pcaps + glob('flows/brex/*')
    if len(pcaps) == 0: return

    hostPath = getcwd()+'/flows'
    containerPath = '/home/flows'

    # Get statsitics from all pcaps files
    printTerm(f"[LST2.0] ... Converting {len(pcaps)} PCAP Files (it might take several minutes)")
    cicflowmeter = CICFlowMeter('cic', hostPath, containerPath)
    cicflowmeter.instantiate()
    [cicflowmeter.analyze('/home/'+pcap, containerPath) for pcap in pcaps]
    cicflowmeter.delete()

    # Merge all csv files
    printTerm(f"[LST2.0] ... Merging all CSV Files")
    csvs = glob('flows/*.csv')
    csv_content = ''
    def get_content(csv):
        nonlocal csv_content
        with open(csv, 'r') as f:
            csv_content += f.read()
    [get_content(csv) for csv in csvs]

    # Remove duplicate headers
    printTerm(f"[LST2.0] ... Removing Duplicate Headers")
    csv_content = csv_content.split('\n')
    header = csv_content[0]
    csv_content = [content for content in csv_content if not 'Flow ID' in content]
    csv_content = header + '\n' + '\n'.join(csv_content)

    # Save new csv file
    subprocess.run(f'rm flows/dump*', shell=True)
    printTerm(f"[LST2.0] ... Saving CSV File")
    with open(hostPath+'/final_report.csv', 'w') as f:
        f.write(csv_content)


def signal_handler(sig, frame):
    print('You pressed Ctrl+C!')
    unmakeChanges()
    convertPcap()
    sys.exit(0)


def collectLogs():
    printTerm(f"[LST2.0] Collecting Client Logs")
    hosts = ['e1', 'e2', 'm1', 'm2', 'm3', 'm4', 'o1', 'o2', 'd1', 'd2', 'd3', 'd4', 'd5', 'd6', 'd7', 'd8', 'd9', 'd10', 'd11', 'd12', 'd13']
    ips = ['50.3', '50.4', '200.2', '200.3', '200.4', '200.5', '210.2', '210.3', '220.2', '220.3', '220.4', '220.5', '220.6', '220.7', '220.8', '220.9', '220.10', '220.11', '220.12', '220.13', '220.14']
    def getLog(ip, host):
        subprocess.run(f'docker cp {host}:/home/debian/log/192.168.{ip}.log logs/192.168.{ip}.log > /dev/null 2>&1', shell=True)
    [getLog(ip, host) for ip, host in zip(ips, hosts)]


def printTerm(string: str) -> None:
    subprocess.run(f"echo \"{string}\"", shell=True)


# Capture the ctrl+c
signal.signal(signal.SIGINT, signal_handler)

try:
    printTerm("LST 2.0  Copyright (C) 2022\n\n\tThis program comes with ABSOLUTELY NO WARRANTY;\n\tThis is free software, and you are welcome to redistribute it\n\tunder certain conditions;\n\n")
    printTerm("[LST2.0] Iniciando o Experimento")    
    # Create folder to store the 
    subprocess.run("mkdir flows 2>/dev/null", shell=True)

    # Create Bridges and connect them
    createBridge('brint', brint_ip, int_gateway)
    createBridge('brex', brex_ip, ex_gateway)
    nodes['brex'].connect(nodes['brint'])

    subprocess.run(f"ip route add 192.168.200.0/24 dev veth-host-brint", shell=True)
    subprocess.run(f"ip route add 192.168.210.0/24 dev veth-host-brint", shell=True)
    subprocess.run(f"ip route add 192.168.220.0/24 dev veth-host-brint", shell=True)

    # Create seafile server
    nodes['seafile'] = Seafile('seafile')
    nodes['seafile'].instantiate()
    setNetworkConfig(nodes['seafile'], nodes['brint'], external_subnet, 1, setFiles=False)
    nodes['seafile'].updateServerConfig()

    # Create controllers
    createController('c1', 'brint', c1_ip, c1port)
    createController('c2', 'brex', c1_ip, c1port)

    # Create server subnet
    createServer('mail',   mailserver,   server_subnet, 1)
    createServer('file',   fileserver,   server_subnet, 2)
    createServer('web',    webserver,    server_subnet, 3)
    createServer('backup', backupserver, server_subnet, 4)

    # Set Management Subnet
    createPrinter('mprinter', management_subnet)
    createLinuxClient('m1', nodes['brint'], management_subnet, 2)
    createLinuxClient('m2', nodes['brint'], management_subnet, 3)
    createLinuxClient('m3', nodes['brint'], management_subnet, 4)
    createLinuxClient('m4', nodes['brint'], management_subnet, 5)
        
    # Set Office Subnet
    createPrinter('oprinter', office_subnet)
    createLinuxClient('o1', nodes['brint'], office_subnet, 2)
    createLinuxClient('o2', nodes['brint'], office_subnet, 3)

    # Set Developer Subnet
    createPrinter('dprinter', developer_subnet)
    createLinuxClient('d1', nodes['brint'], developer_subnet, 2)
    createLinuxClient('d2', nodes['brint'], developer_subnet, 3)
    createLinuxClient('d3', nodes['brint'], developer_subnet, 4)
    createLinuxClient('d4', nodes['brint'], developer_subnet, 5)
    createLinuxClient('d5', nodes['brint'], developer_subnet, 6)
    createLinuxClient('d6', nodes['brint'], developer_subnet, 7)
    createLinuxClient('d7', nodes['brint'], developer_subnet, 8)
    createLinuxClient('d8', nodes['brint'], developer_subnet, 9)
    createLinuxClient('d9', nodes['brint'], developer_subnet, 10)
    createLinuxClient('d10', nodes['brint'], developer_subnet, 11)
    createLinuxClient('d11', nodes['brint'], developer_subnet, 12)
    createLinuxClient('d12', nodes['brint'], developer_subnet, 13)
    createLinuxClient('d13', nodes['brint'], developer_subnet, 14)

    # Set External Subnet
    createServer('eweb', webserver, external_subnet, 2)
    createLinuxClient('e1', nodes['brex'], external_subnet, 3)
    createLinuxClient('e2', nodes['brex'], external_subnet, 4)

    # Set Configuration Files
    [setLinuxClientFileConfig(nodes[f'm{i}'], management_subnet, 'management') for i in range(1, 5)]
    [setLinuxClientFileConfig(nodes[f'o{i}'], office_subnet, 'office') for i in range(1,3)]
    [setLinuxClientFileConfig(nodes[f'd{i}'], developer_subnet, 'administrator') for i in range(1,3)]
    [setLinuxClientFileConfig(nodes[f'd{i}'], developer_subnet, 'developer') for i in range(3,12)]
    [setLinuxClientFileConfig(nodes[f'd{i}'], developer_subnet, 'attacker') for i in range(12,14)]
    [setLinuxClientFileConfig(nodes[f'e{i}'], developer_subnet, 'external_attacker') for i in range(1,3)]

    nodes['brint'].collectFlows(path='home/pcap', sniffAll=True)
    nodes['brex'].collectFlows(path='home/pcap', sniffAll=True)
except Exception as ex:
    collectLogs()
    unmakeChanges()
    convertPcap()
    raise Exception(str(ex))

print('[LST2.0] Press CTRL+C to finish experiment')
signal.pause()



from .pschedulerWrapper import Throughput, Rtt, Latency
from .constants import *


throughput = Throughput().Format(PERFSONAR_JSON_OUTPUT_FORMAT).MaxRuns(MAX_RUNS).Repeat(INTERVAL)
rtt = Rtt().Format(PERFSONAR_JSON_OUTPUT_FORMAT).MaxRuns(25).Repeat(REPEAT_INTERVAL)
latency = Latency().Format(PERFSONAR_JSON_OUTPUT_FORMAT).MaxRuns(MAX_RUNS).Repeat(INTERVAL)


def runThroughput(testname, sourceIp, targetIp):
        throughput.Source(sourceIp)\
                .Dest(targetIp)\
                .OutputFile(RESULTS_PATH, testname + THROUGHPUT_JSON_FORMAT)\
                .ThroughputDuration(60)\
                .mountCommand()
        print("Running now command " + throughput.command)
        throughput.run()


def runRTT(testname, sourceIp, targetIp):
        rtt.Source(sourceIp)\
                .Dest(targetIp)\
                .OutputFile(RESULTS_PATH, testname + RTT_JSON_FORMAT)\
                .Count(60)\
                .mountCommand()
        print("Running now command " + rtt.command)
        rtt.run()


def runLatency(testname, sourceIp, targetIp):
        latency.Source(sourceIp)\
                .Dest(targetIp)\
                .OutputRaw()\
                .OutputFile(RESULTS_PATH, testname + LATENCY_JSON_FORMAT)\
                .PacketCount(60)\
                .mountCommand()
        print("Running now command " + latency.command)
        latency.run()
        



# Copyright (C) 2022 Alexandre Mitsuru Kaihara
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import subprocess
from exceptions import *


# Just to enable the declaration of Type in methods
class Node:
    pass


# Brief: This is a super class that define methods common for network nodes
class Node:
    # Brief: Constructor of Node super class
    # Params:
    #   String containerName: Name of the container
    # Return:
    #   None
    def __init__(self, nodeName: str) -> None:
        self.__nodeName = nodeName

    # OBS: Create nodes with short name lenght due to a restriction on a iproute2 to define and create interfaces. Peers names can't have more than 14 chars
    # Brief: Instantiate the container
    # Params:
    #   String dockerImage: Name of the container of a local image or in a Docker Hub repository
    #   String DockerCommand: String to be used to instantiate the container instead of the standard command
    # Return:
    #   None
    def instantiate(self, dockerImage="alexandremitsurukaihara/lst2.0:host", dockerCommand = '', dns='8.8.8.8') -> None:
        if not self.__imageExists(dockerImage):
            self.__pullImage(dockerImage)
        try:    
            if dockerCommand == '':
                subprocess.run(f"docker run -d --network=none --privileged --name={self.getNodeName()} --dns={dns} {dockerImage}", shell=True, capture_output=True)
            else:
                subprocess.run(dockerCommand, shell=True, capture_output=True)
        except Exception as ex:
            logging.error(f"Error while criating the container {self.getNodeName()}: {str(ex)}")
            raise NodeInstantiationFailed(f"Error while criating the container {self.getNodeName()}: {str(ex)}")
        self.__enableNamespace(self.getNodeName())

    # Brief: Verifies if the image exists
    # Params:
    #   String image: Tag of the Docker image 
    # Return:
    #   True if the image exists locally
    def __imageExists(self, image: str) -> bool:
        out = self.run(f"docker inspect --type=image {image}")
        if '[]' in out.stdout.decode('utf-8'): return False
        else: return True
            
    # Brief: Pulls the image from a Docker Hub repository
    # Params:
    #   String image: Tag of the Docker image 
    # Return:
    #   True if the image exists locally
    def __pullImage(self, image):
        try: 
            subprocess.run(f"docker pull {image}", shell=True)
        except Exception as ex:
            logging.error(f"Error pulling non-existing {image} image: {str(ex)}")
            raise NodeInstantiationFailed(f"Error pulling non-existing {image} image: {str(ex)}")

    # Brief: Instantiate the container
    # Params:
    #   String dockerImage: Name of the container of a local image or in a Docker Hub repository
    #   String DockerCommand: String to be used to instantiate the container instead of the standard command
    # Return:
    #   None
    def delete(self) -> None:
        try:    
            subprocess.run(f"docker kill {self.getNodeName()} && docker rm {self.getNodeName()}", shell=True, capture_output=True)
        except Exception as ex:
            logging.error(f"Error while deleting the host {self.getNodeName()}: {str(ex)}")
            raise NodeInstantiationFailed(f"Error while deleting the host {self.getNodeName()}: {str(ex)}")

    # Brief: Set Ip to an interface (the ip must be set only after connecting it to a container)
    # Params:
    #   String ip: IP address to be set to peerName interface
    #   int mask: Integer that represents the network mask
    #   String node: Reference to the node it is connected to this container to discover the intereface to set the ip to
    # Return:
    #   None
    def setIp(self, ip: str, mask: int, node: Node) -> None:
        if not self.__isConnected(node):
            logging.error(f"Incorrect node reference for {node.getNodeName()}, connect {self.getNodeName()} first")
            raise Exception(f"Incorrect node reference for {node.getNodeName()}, connect {self.getNodeName()} first")

        interfaceName = self.__getThisInterfaceName(node)
        self.__setIp(ip, mask, interfaceName)

    # Brief: Creates Linux virtual interfaces and connects peers to the nodes, in case of one of the nodes is a switch, it also creates a port in bridge
    # Params:
    #   Node node: Reference of another node to connect to
    # Return:
    #   None
    def connect(self, node: Node) -> None:
        if self.__isConnected(node):
            logging.error(f"Cannot connect to {node.getNodeName()}, node already connected")
            raise Exception(f"Cannot connect to {node.getNodeName()}, node already connected")

        peer1Name = self.__getThisInterfaceName(node)
        peer2Name = self.__getOtherInterfaceName(node)
        
        self.__create(peer1Name, peer2Name)
        self.__setInterface(self.getNodeName(), peer1Name)
        self.__setInterface(node.getNodeName(), peer2Name)

        if self.__class__.__name__ == 'Switch':
            self._Switch__createPort(self.getNodeName(), self.__getThisInterfaceName(node))
        if node.__class__.__name__ == 'Switch':
            node._Switch__createPort(node.getNodeName(), node.__getThisInterfaceName(self))

    
    def connectToInternet(self, hostIP: str, hostMask: int) -> None:
        peer1Name = f"veth-{self.getNodeName()}-host"
        peer2Name = f"veth-host-{self.getNodeName()}"

        self.__create(peer1Name, peer2Name)
        self.__setInterface(self.getNodeName(), peer1Name)
        if self.__class__.__name__ == 'Switch':
            self._Switch__createPort(self.getNodeName(), peer1Name)
        
        subprocess.run(f"ip link set {peer2Name} up", shell=True)
        subprocess.run(f"ip addr add {hostIP}/{hostMask} dev {peer2Name}", shell=True)

        # Enable forwading packets from host to interface
        hostGatewayInterface = subprocess.run(f"route | grep \'^default' | grep -o '[^ ]*$\'", shell=True, capture_output=True).stdout.decode('utf8').replace("\n", '')
        subprocess.run(f"iptables -t nat -I POSTROUTING -o {hostGatewayInterface} -j MASQUERADE", shell=True)
        subprocess.run(f"iptables -t nat -I POSTROUTING -o {peer2Name} -j MASQUERADE", shell=True)
        subprocess.run(f"iptables -A FORWARD -i {peer2Name} -o {hostGatewayInterface} -j ACCEPT", shell=True)
        subprocess.run(f"iptables -A FORWARD -i {hostGatewayInterface} -o {peer2Name} -j ACCEPT", shell=True)

    # Brief: Returns the value of the container name
    # Params:
    # Return:
    #   Returns the name of the node
    def getNodeName(self) -> str:
        return self.__nodeName

    # Brief: Add a route in routing table of container
    # Params:
    #   String ip: IP address of the route
    #   String mask: Network mask for the IP address route
    #   String interfaceName: Name of the interface to forward to
    # Return:
    #   None
    def addRoute(self, ip: str, mask: int,  node: Node):
        if not self.__isConnected(node):
            logging.error(f"Incorrect node reference for {node.getNodeName()}, connect {self.getNodeName()} first")
            raise Exception(f"Incorrect node reference for {node.getNodeName()}, connect {self.getNodeName()} first")
        peerName = self.__getThisInterfaceName(node)
        try:
            subprocess.run(f"docker exec {self.getNodeName()} ip route add {ip}/{mask} dev {peerName}", shell=True)
        except Exception as ex:
            logging.error(f"Error adding route {ip}/{mask} via {peerName} in {self.getNodeName()}: {str(ex)}")
            raise Exception(f"Error adding route {ip}/{mask} via {peerName} in {self.getNodeName()}: {str(ex)}")

    # Brief: Set Ip to an interface (the ip must be set only after connecting it to a container, because)
    # Params:
    #   String destinationIp: The destination IP address of the gateway in format "XXX.XXX.XXX.XXX"
    #   String node: Reference to node that will serve as the first hop to forward the packets to the gateway, this reference node must be already connected to it
    # Return:
    #   None
    def setDefaultGateway(self, destinationIp: str, node: Node) -> None:
        if not self.__isConnected(node):
            logging.error(f"Incorrect node reference for {node.getNodeName()}, connect {self.getNodeName()} first")
            raise Exception(f"Incorrect node reference for {node.getNodeName()}, connect {self.getNodeName()} first")
        
        self.addRoute(destinationIp, 32, node)
        outputInterface = self.__getThisInterfaceName(node)
        try:
            subprocess.run(f"docker exec {self.getNodeName()} route add default gw {destinationIp} dev {outputInterface}", shell=True)
        except Exception as ex:
            logging.error(f"Error while setting gateway {destinationIp} on device {outputInterface} in {self.getNodeName()}: {str(ex)}")
            raise Exception(f"Error while setting gateway {destinationIp} on device {outputInterface} in {self.getNodeName()}: {str(ex)}")

    # Brief: Runs a command inside the container
    # Params:
    #   String command: String containing the command to run inside the container
    # Return:
    #   Returns variable that contains stdout and stderr (more information in subprocess documentation)
    def run(self, command: str) -> str:
        try:
            return subprocess.run(f'docker exec {self.getNodeName()} '+command, shell=True, capture_output=True)
        except Exception as ex:
            logging.error(f"Error executing command {command} in {self.getNodeName()}: {str(ex)}")
            raise Exception(f"Error executing command {command} in {self.getNodeName()}: {str(ex)}")

    # Brief: Runs multiple commands inside the container
    # Params:
    #   List<String> commands: Runs multiple comands inside the container
    # Return:
    #   Returns a list with the variable that contains stdout and stderr (more information in subprocess documentation)
    def runs(self, commands: list) -> list:
        return [self.run(command) for command in commands]

    # Brief: Copy local file into container
    # Params:
    #   String path: Absolute or relative path to the file to be copied from local (path+filename)
    #   String destPath: Absolute path to copy the file to the container (path+filename)
    # Return:
    def copyLocalToContainer(self, path: str, destPath: str) -> None:
        try:
            subprocess.run(f"docker cp {path} {self.getNodeName()}:{destPath}", shell=True, capture_output=True)
        except Exception as ex:
            logging.error(f"Error copying file from {path} to {destPath}: {str(ex)}")
            raise Exception(f"Error copying file from {path} to {destPath}: {str(ex)}")

    # Brief: Copy local file into container
    # Params:
    #   String path: Absolute path to the file to be copied from container (path+filename)
    #   String destPath: Absolute or relative path to copy to local (path+filename)
    # Return:
    def copyContainerToLocal(self, path: str, destPath: str) -> None:
        try:
            subprocess.run(f"docker cp {self.getNodeName()}:{path} {destPath}", shell=True, capture_output=True)
        except Exception as ex:
            logging.error(f"Error copying file from {path} to {destPath}: {str(ex)}")
            raise Exception(f"Error copying file from {path} to {destPath}: {str(ex)}")

    # Brief: Returns the name of the interface to be created on this node
    # Params:
    #   Node node: Reference of another node to connect to
    # Return:
    #   Name of the interface with pattern veth + this node name + other node name
    def __getThisInterfaceName(self, node: Node) -> str:
        return self.getNodeName()+node.getNodeName()

    # Brief: Set Ip to an interface
    # Params:
    #   String ip: IP address to be set to peerName interface
    #   int mask: Integer that represents the network mask
    #   String interfaceName: Name of the interface to set the ip
    # Return:
    #   None
    def __setIp(self, ip: str, mask: int, interfaceName: str) -> None:
        try:
            subprocess.run(f"ip -n {self.getNodeName()} addr add {ip}/{mask} dev {interfaceName}", shell=True)
        except Exception as ex:
            logging.error(f"Error while setting IP {ip}/{mask} to virtual interface {interfaceName}: {str(ex)}")
            raise Exception(f"Error while setting IP {ip}/{mask} to virtual interface {interfaceName}: {str(ex)}")

    # Brief: Returns the name of the interface to be created on other node
    # Params:
    #   Node node: Reference of another node to connect to
    # Return:
    #   Name of the interface with pattern veth + other node name + this node name
    def __getOtherInterfaceName(self, node: Node) -> str:
        return node.getNodeName()+self.getNodeName()

    # Brief: Creates the virtual interfaces and set them up (names cant be the same as some existing one in host's namespace)
    # Params:
    #   String peer1Name: Name of the interface to connect to the first peer 
    #   String peer2Name: Name of the interface to connect to the second peer 
    # Return:
    #   None
    def __create(self, peer1Name: str, peer2Name: str) -> None:
        try:
            subprocess.run(f"ip link add {peer1Name} type veth peer name {peer2Name}", shell=True)
        except Exception as ex:
            logging.error(f"Error while creating virtual interfaces {peer1Name} and {peer2Name}: {str(ex)}")
            raise Exception(f"Error while creating virtual interfaces {peer1Name} and {peer2Name}: {str(ex)}")

    # Brief: Set the interface to node
    # Params:
    #   String nodeName: Name of the node network namespace
    #   String peerName: Name of the interface to set to node
    # Return:
    #   None
    def __setInterface(self, nodeName: str, peerName: str) -> None:
        try:
            subprocess.run(f"ip link set {peerName} netns {nodeName}", shell=True)
            subprocess.run(f"ip -n {nodeName} link set {peerName} up", shell=True)
        except Exception as ex:
            logging.error(f"Error while setting virtual interfaces {peerName} to {nodeName}: {str(ex)}")
            raise Exception(f"Error while setting virtual interfaces {peerName} to {nodeName}: {str(ex)}")

    # Brief: Enable accessing the Docker node namespace directly
    # Params:
    # Return:
    #   None
    def __enableNamespace(self, nodeName) -> None:
        try:    
            subprocess.run(f"pid=$(docker inspect -f '{{{{.State.Pid}}}}' {nodeName}); mkdir -p /var/run/netns/; ln -sfT /proc/$pid/ns/net /var/run/netns/{nodeName}", shell=True)
        except Exception as ex:
            logging.error(f"Error while deleting the host {self.getNodeName()}: {str(ex)}")
            raise Exception(f"Error while deleting the host {self.getNodeName()}: {str(ex)}")

    # Brief: Verifies if there is a connection with a node
    # Params:
    #   Node node: Node to check connection
    # Return:
    #   Return true if it is connected or false otherwise
    def __isConnected(self, node: Node) -> bool:
        interfaceName = self.__getThisInterfaceName(node)
        interfaces = self.__getAllIntefaces()
        for interface in interfaces: 
            if interface == interfaceName: return True
        return False

    # Brief: Get all interfaces names
    # Params:
    # Return:
    #   Return a list with the name of all interfaces
    def __getAllIntefaces(self) -> list:
        output = subprocess.run(f"docker exec {self.getNodeName()} ifconfig -a | sed 's/[ \t].*//;/^$/d'", shell=True, capture_output=True)
        interfaces=output.stdout.decode('utf8').replace(":", '').split('\n')
        return list(filter(None, interfaces)) # Remove empty strings

    # Brief: Verifies if the container is active
    # Params:
    # Return:
    #   Return true if it is active or false otherwise
    def __isActive(self) -> bool:
        if subprocess.run(f"docker ps | grep {self.getNodeName()}'", shell=True, capture_output=True).stdout.decode('utf8') != '': return True
        return False



"""
Node objects for Mininet.

Nodes provide a simple abstraction for interacting with hosts, switches
and controllers. Local nodes are simply one or more processes on the local
machine.

Node: superclass for all (primarily local) network nodes.

Host: a virtual host. By default, a host is simply a shell; commands
    may be sent using Cmd (which waits for output), or using sendCmd(),
    which returns immediately, allowing subsequent monitoring using
    monitor(). Examples of how to run experiments using this
    functionality are provided in the examples/ directory. By default,
    hosts share the root file system, but they may also specify private
    directories.

CPULimitedHost: a virtual host whose CPU bandwidth is limited by
    RT or CFS bandwidth limiting.

Switch: superclass for switch nodes.

UserSwitch: a switch using the user-space switch from the OpenFlow
    reference implementation.

OVSSwitch: a switch using the Open vSwitch OpenFlow-compatible switch
    implementation (openvswitch.org).

Controller: superclass for OpenFlow controllers. The default controller
    is controller(8) from the reference implementation.

RemoteController: a remote controller node, which may use any
    arbitrary OpenFlow-compatible controller, and which is not
    created or managed by Mininet.

Future enhancements:

- Possibly make Node, Switch and Controller more abstract so that
  they can be used for both local and remote nodes

- Create proxy objects for remote nodes (Mininet: Cluster Edition)
"""
import errno
import os
import re
import pty
import select
import docker
import json
from subprocess import check_output
from time import sleep
from re import findall

from mininet.node import Node, Host
from mininet.moduledeps import moduleDeps, pathCheck, TUN
from mininet.util import ( quietRun, errRun, which)
from mininet.log import info, error, warn, debug
from mn_wifi.node import Station
from containernet.link import Intf, TCIntf, OVSIntf
from distutils.version import StrictVersion


class Docker ( Host ):
    """Node that represents a docker container.
    This part is inspired by:
    http://techandtrains.com/2014/08/21/docker-container-as-mininet-host/
    We use the docker-py client library to control docker.
    """

    def __init__(self, name, dimage=None, dcmd=None, did=None, **kwargs):
        """
        Creates a Docker container as Mininet host.

        Resource limitations based on CFS scheduler:
        * cpu.cfs_quota_us: the total available run-time within a period (in microseconds)
        * cpu.cfs_period_us: the length of a period (in microseconds)
        (https://www.kernel.org/doc/Documentation/scheduler/sched-bwc.txt)

        Default Docker resource limitations:
        * cpu_shares: Relative amount of max. avail CPU for container
            (not a hard limit, e.g. if only one container is busy and the rest idle)
            e.g. usage: d1=4 d2=6 <=> 40% 60% CPU
        * cpuset_cpus: Bind containers to CPU 0 = cpu_1 ... n-1 = cpu_n (string: '0,2')
        * mem_limit: Memory limit (format: <number>[<unit>], where unit = b, k, m or g)
        * memswap_limit: Total limit = memory + swap

        All resource limits can be updated at runtime! Use:
        * updateCpuLimits(...)
        * updateMemoryLimits(...)
        """
        self.dimage = dimage
        self.dnameprefix = "mn"
        self.dcmd = dcmd if dcmd is not None else "/bin/bash"
        self.dc = None  # pointer to the dict containing 'Id' and 'Warnings' keys of the container
        self.dcinfo = None
        self.func = []
        self.did = did  # Id of running container
        #  let's store our resource limits to have them available through the
        #  Mininet API later on
        defaults = { 'cpu_quota': -1,
                     'cpu_period': None,
                     'cpu_shares': None,
                     'cpuset_cpus': None,
                     'mem_limit': None,
                     'memswap_limit': None,
                     'environment': {},
                     'volumes': [],  # use ["/home/user1/:/mnt/vol2:rw"]
                     'tmpfs': [],  # use ["/home/vol1/:size=3G,uid=1000"]
                     'network_mode': None,
                     'privileged': False,
                     'publish_all_ports': True,
                     'port_bindings': {},
                     'ports': [],
                     'dns': [],
                     'ipc_mode': None,
                     'devices': [],
                     'cap_add': ['net_admin'],  # we need this to allow mininet network setup
                     #'storage_opt': None,
                     'sysctls': {}
                     }
        defaults.update( kwargs )

        if 'net_admin' not in defaults['cap_add']:
            defaults['cap_add'] += ['net_admin']  # adding net_admin if it's cleared out to allow mininet network setup

        # keep resource in a dict for easy update during container lifetime
        self.resources = dict(
            cpu_quota=defaults['cpu_quota'],
            cpu_period=defaults['cpu_period'],
            cpu_shares=defaults['cpu_shares'],
            cpuset_cpus=defaults['cpuset_cpus'],
            mem_limit=defaults['mem_limit'],
            memswap_limit=defaults['memswap_limit']
        )

        self.volumes = defaults['volumes']
        self.tmpfs = defaults['tmpfs']
        self.environment = {} if defaults['environment'] is None else defaults['environment']
        # setting PS1 at "docker run" may break the python docker api (update_container hangs...)
        # self.environment.update({"PS1": chr(127)})  # CLI support
        self.network_mode = defaults['network_mode']
        self.privileged = defaults['privileged']
        self.publish_all_ports = defaults['publish_all_ports']
        self.port_bindings = defaults['port_bindings']
        self.dns = defaults['dns']
        self.ipc_mode = defaults['ipc_mode']
        self.devices = defaults['devices']
        self.cap_add = defaults['cap_add']
        self.sysctls = defaults['sysctls']
        #self.storage_opt = defaults['storage_opt']

        # setup docker client
        # self.dcli = docker.APIClient(base_url='unix://var/run/docker.sock')
        self.dcli = docker.from_env().api

        # pull image if it does not exist
        if self.dimage:
            self._check_image_exists(dimage, True)

        # for DEBUG
        debug("Created docker container object %s\n" % name)
        debug("image: %s\n" % str(self.dimage))
        debug("dcmd: %s\n" % str(self.dcmd))
        info("%s: kwargs %s\n" % (name, str(kwargs)))

        # creats host config for container
        # see: https://docker-py.readthedocs.org/en/latest/hostconfig/
        hc = self.dcli.create_host_config(
            network_mode=self.network_mode,
            privileged=self.privileged,
            binds=self.volumes,
            tmpfs=self.tmpfs,
            publish_all_ports=self.publish_all_ports,
            port_bindings=self.port_bindings,
            mem_limit=self.resources.get('mem_limit'),
            cpuset_cpus=self.resources.get('cpuset_cpus'),
            dns=self.dns,
            ipc_mode=self.ipc_mode,  # string
            devices=self.devices,  # see docker-py docu
            cap_add=self.cap_add,  # see docker-py docu
            sysctls=self.sysctls,  # see docker-py docu
            #storage_opt=self.storage_opt
        )

        if kwargs.get("rm", False):
            container_list = self.dcli.containers(all=True)
            for container in container_list:
                for container_name in container.get("Names", []):
                    if "%s.%s" % (self.dnameprefix, name) in container_name:
                        self.dcli.remove_container(container="%s.%s" % (self.dnameprefix, name), force=True)
                        break

        if self.did:
            self.dc = {'Id': '{}'.format(self.did), 'Warnings': []}
            self.existing_container = 1
        else:
            # create new docker container
            self.dc = self.dcli.create_container(
                name="%s.%s" % (self.dnameprefix, name),
                image=self.dimage,
                command=self.dcmd,
                entrypoint=list(),  # overwrite (will be executed manually at the end)
                stdin_open=True,  # keep container open
                tty=True,  # allocate pseudo tty
                environment=self.environment,
                #network_disabled=True,  # docker stats breaks if we disable the default network
                host_config=hc,
                ports=defaults['ports'],
                labels=['com.containernet'],
                volumes=[self._get_volume_mount_name(v) for v in self.volumes if self._get_volume_mount_name(v) is not None],
                hostname=name
            )

        # start the container
        self.dcli.start(self.dc)
        debug("Docker container %s started\n" % name)

        # fetch information about new container
        self.dcinfo = self.dcli.inspect_container(self.dc)
        self.did = self.dcinfo.get("Id")

        # call original Node.__init__
        Node.__init__(self, name, **kwargs)

        # let's initially set our resource limits
        self.update_resources(**self.resources)

    def start(self):
        # Containernet ignores the CMD field of the Dockerfile.
        # Lets try to load it here and manually execute it once the
        # container is started and configured by Containernet:
        cmd_field = self.get_cmd_field(self.dimage)
        entryp_field = self.get_entrypoint_field(self.dimage)
        if entryp_field is not None:
            if cmd_field is None:
                cmd_field = list()
            # clean up cmd_field
            try:
                cmd_field.remove(u'/bin/sh')
                cmd_field.remove(u'-c')
            except ValueError:
                pass
            # we just add the entryp. commands to the beginning:
            cmd_field = entryp_field + cmd_field
        if cmd_field is not None:
            cmd_field.append("> /dev/pts/0 2>&1")  # make output available to docker logs
            cmd_field.append("&")  # put to background (works, but not nice)
            info("{}: running CMD: {}\n".format(self.name, cmd_field))
            self.cmd(" ".join(cmd_field))

    def get_cmd_field(self, imagename):
        """
        Try to find the original CMD command of the Dockerfile
        by inspecting the Docker image.
        Returns list from CMD field if it is different from
        a single /bin/bash command which Containernet executes
        anyhow.
        """
        try:
            imgd = self.dcli.inspect_image(imagename)
            cmd = imgd.get("Config", {}).get("Cmd")
            assert isinstance(cmd, list)
            # filter the default case: a single "/bin/bash"
            if "/bin/bash" in cmd and len(cmd) == 1:
                return None
            return cmd
        except BaseException as ex:
            error("Error during image inspection of {}:{}"
                  .format(imagename, ex))
        return None

    def get_entrypoint_field(self, imagename):
        """
        Try to find the original ENTRYPOINT command of the Dockerfile
        by inspecting the Docker image.
        Returns list or None.
        """
        try:
            imgd = self.dcli.inspect_image(imagename)
            ep = imgd.get("Config", {}).get("Entrypoint")
            if isinstance(ep, list) and len(ep) < 1:
                return None
            return ep
        except BaseException as ex:
            error("Error during image inspection of {}:{}"
                  .format(imagename, ex))
        return None

    # Command support via shell process in namespace
    def startShell( self, *args, **kwargs ):
        "Start a shell process for running commands"
        if self.shell:
            error( "%s: shell is already running\n" % self.name )
            return
        # mnexec: (c)lose descriptors, (d)etach from tty,
        # (p)rint pid, and run in (n)amespace
        # opts = '-cd' if mnopts is None else mnopts
        # if self.inNamespace:
        #     opts += 'n'
        # bash -i: force interactive
        # -s: pass $* to shell, and make process easy to find in ps
        # prompt is set to sentinel chr( 127 )
        if hasattr(self, 'existing_container'):
            cmd = ['docker', 'exec', '-it', '%s' % self.did, 'env', 'PS1=' + chr(127),
                   'bash', '--norc', '-is', 'mininet:' + self.name]
        else:
            cmd = ['docker', 'exec', '-it', '%s.%s' % (self.dnameprefix, self.name), 'env', 'PS1=' + chr(127),
                   'bash', '--norc', '-is', 'mininet:' + self.name]
        # Spawn a shell subprocess in a pseudo-tty, to disable buffering
        # in the subprocess and insulate it from signals (e.g. SIGINT)
        # received by the parent
        self.master, self.slave = pty.openpty()
        self.shell = self._popen( cmd, stdin=self.slave, stdout=self.slave, stderr=self.slave,
                                  close_fds=False )
        self.stdin = os.fdopen( self.master, 'r' )
        self.stdout = self.stdin
        self.pid = self._get_pid()
        self.pollOut = select.poll()
        self.pollOut.register( self.stdout )
        # Maintain mapping between file descriptors and nodes
        # This is useful for monitoring multiple nodes
        # using select.poll()
        self.outToNode[ self.stdout.fileno() ] = self
        self.inToNode[ self.stdin.fileno() ] = self
        self.execed = False
        self.lastCmd = None
        self.lastPid = None
        self.readbuf = ''
        # Wait for prompt
        while True:
            data = self.read( 1024 )
            if data[ -1 ] == chr( 127 ):
                break
            self.pollOut.poll()
        self.waiting = False
        # +m: disable job control notification
        self.cmd( 'unset HISTFILE; stty -echo; set +m' )

    def _get_volume_mount_name(self, volume_str):
        """ Helper to extract mount names from volume specification strings """
        parts = volume_str.split(":")
        if len(parts) < 3:
            return None
        return parts[1]

    def terminate( self ):
        """ Stop docker container """
        if not self._is_container_running():
            return
        try:
            if not hasattr(self, 'existing_container'):
                self.dcli.remove_container(self.dc, force=True, v=True)
        except docker.errors.APIError:
            warn("Warning: API error during container removal.\n")

        self.cleanup()

    def sendCmd( self, *args, **kwargs ):
        """Send a command, followed by a command to echo a sentinel,
           and return without waiting for the command to complete."""
        self._check_shell()
        if not self.shell:
            return
        Host.sendCmd( self, *args, **kwargs )

    def popen( self, *args, **kwargs ):
        """Return a Popen() object in node's namespace
           args: Popen() args, single list, or string
           kwargs: Popen() keyword args"""
        if not self._is_container_running():
            error( "ERROR: Can't connect to Container \'%s\'' for docker host \'%s\'!\n" % (self.did, self.name) )
            return
        mncmd = ["docker", "exec", "-t", "%s.%s" % (self.dnameprefix, self.name)]
        return Node.popen( self, *args, mncmd=mncmd, **kwargs )

    def cmd(self, *args, **kwargs ):
        """Send a command, wait for output, and return it.
           cmd: string"""
        verbose = kwargs.get( 'verbose', False )
        log = info if verbose else debug
        log( '*** %s : %s\n' % ( self.name, args ) )
        self.sendCmd( *args, **kwargs )
        return self.waitOutput( verbose )

    def _get_pid(self):
        state = self.dcinfo.get("State", None)
        if state:
            return state.get("Pid", -1)
        return -1

    def _check_shell(self):
        """Verify if shell is alive and
           try to restart if needed"""
        if self._is_container_running():
            if self.shell:
                self.shell.poll()
                if self.shell.returncode is not None:
                    debug("*** Shell died for docker host \'%s\'!\n" % self.name )
                    self.shell = None
                    debug("*** Restarting Shell of docker host \'%s\'!\n" % self.name )
                    self.startShell()
            else:
                debug("*** Restarting Shell of docker host \'%s\'!\n" % self.name )
                self.startShell()
        else:
            error( "ERROR: Can't connect to Container \'%s\'' for docker host \'%s\'!\n" % (self.did, self.name) )
            if self.shell:
                self.shell = None

    def _is_container_running(self):
        """Verify if container is alive"""
        container_list = self.dcli.containers(filters={"id": self.did, "status": "running"})
        if len(container_list) == 0:
            return False;
        return True

    def _check_image_exists(self, imagename, pullImage=False):
        # split tag from repository if a tag is specified
        if ":" in imagename:
            #If two :, then the first is to specify a port. Otherwise, it must be a tag
            slices = imagename.split(":")
            repo = ":".join(slices[0:-1])
            tag = slices[-1]
        else:
            repo = imagename
            tag = "latest"

        if self._image_exists(repo, tag):
            return True

        # image not found
        if pullImage:
            if self._pull_image(repo, tag):
                info('*** Download of "%s:%s" successful\n' % (repo, tag))
                return True
        # we couldn't find the image
        return False

    def _image_exists(self, repo, tag):
        """
        Checks if the repo:tag image exists locally
        :return: True if the image exists locally. Else false.
        """
        # filter by repository
        images = self.dcli.images(repo)
        imageName = "%s:%s" % (repo, tag)

        for image in images:
            if 'RepoTags' in image:
                if image['RepoTags'] is None:
                    return False
                if imageName in image['RepoTags']:
                    return True
        return False

    def _pull_image(self, repository, tag):
        """
        :return: True if pull was successful. Else false.
        """
        try:
            info('*** Image "%s:%s" not found. Trying to load the image. \n' % (repository, tag))
            info('*** This can take some minutes...\n')

            message = ""
            for line in self.dcli.pull(repository, tag, stream=True):
                # Collect output of the log for enhanced error feedback
                message = message + json.dumps(json.loads(line), indent=4)

        except:
            error('*** error: _pull_image: %s:%s failed.' % (repository, tag)
                  + message)
        if not self._image_exists(repository, tag):
            error('*** error: _pull_image: %s:%s failed.' % (repository, tag)
                  + message)
            return False
        return True

    def update_resources(self, **kwargs):
        """
        Update the container's resources using the docker.update function
        re-using the same parameters:
        Args:
           blkio_weight
           cpu_period, cpu_quota, cpu_shares
           cpuset_cpus
           cpuset_mems
           mem_limit
           mem_reservation
           memswap_limit
           kernel_memory
           restart_policy
        see https://docs.docker.com/engine/reference/commandline/update/
        or API docs: https://docker-py.readthedocs.io/en/stable/api.html#module-docker.api.container
        :return:
        """

        self.resources.update(kwargs)
        # filter out None values to avoid errors
        resources_filtered = {res:self.resources[res] for res in self.resources if self.resources[res] is not None}
        info("{1}: update resources {0}\n".format(resources_filtered, self.name))
        self.dcli.update_container(self.dc, **resources_filtered)

    def updateCpuLimit(self, cpu_quota=-1, cpu_period=-1, cpu_shares=-1, cores=None):
        """
        Update CPU resource limitations.
        This method allows to update resource limitations at runtime by bypassing the Docker API
        and directly manipulating the cgroup options.
        Args:
            cpu_quota: cfs quota us
            cpu_period: cfs period us
            cpu_shares: cpu shares
            cores: specifies which cores should be accessible for the container e.g. "0-2,16" represents
                Cores 0, 1, 2, and 16
        """
        # see https://www.kernel.org/doc/Documentation/scheduler/sched-bwc.txt

        # also negative values can be set for cpu_quota (uncontrained setting)
        # just check if value is a valid integer
        if isinstance(cpu_quota, int):
            self.resources['cpu_quota'] = self.cgroupSet("cfs_quota_us", cpu_quota)
        if cpu_period >= 0:
            self.resources['cpu_period'] = self.cgroupSet("cfs_period_us", cpu_period)
        if cpu_shares >= 0:
            self.resources['cpu_shares'] = self.cgroupSet("shares", cpu_shares)
        if cores:
            self.dcli.update_container(self.dc, cpuset_cpus=cores)
            # quota, period ad shares can also be set by this line. Usable for future work.

    def updateMemoryLimit(self, mem_limit=-1, memswap_limit=-1):
        """
        Update Memory resource limitations.
        This method allows to update resource limitations at runtime by bypassing the Docker API
        and directly manipulating the cgroup options.

        Args:
            mem_limit: memory limit in bytes
            memswap_limit: swap limit in bytes

        """
        # see https://www.kernel.org/doc/Documentation/scheduler/sched-bwc.txt
        if mem_limit >= 0:
            self.resources['mem_limit'] = self.cgroupSet("limit_in_bytes", mem_limit, resource="memory")
        if memswap_limit >= 0:
            self.resources['memswap_limit'] = self.cgroupSet("memsw.limit_in_bytes", memswap_limit, resource="memory")


    def cgroupSet(self, param, value, resource='cpu'):
        """
        Directly manipulate the resource settings of the Docker container's cgrpup.
        Args:
            param: parameter to set, e.g., cfs_quota_us
            value: value to set
            resource: resource name: cpu

        Returns: value that was set

        """
        cmd = 'cgset -r %s.%s=%s docker/%s' % (
            resource, param, value, self.did)
        debug(cmd + "\n")
        try:
            check_output(cmd, shell=True)
        except:
            error("Problem writing cgroup setting %r\n" % cmd)
            return
        nvalue = int(self.cgroupGet(param, resource))
        if nvalue != value:
            error('*** error: cgroupSet: %s set to %s instead of %s\n'
                  % (param, nvalue, value))
        return nvalue

    def cgroupGet(self, param, resource='cpu'):
        """
        Read cgroup values.
        Args:
            param: parameter to read, e.g., cfs_quota_us
            resource: resource name: cpu / memory

        Returns: value

        """
        cmd = 'cgget -r %s.%s docker/%s' % (
            resource, param, self.did)
        try:
            return int(check_output(cmd, shell=True).split()[-1])
        except:
            error("Problem reading cgroup info: %r\n" % cmd)
            return -1


class DockerSta(Station):
    """Node that represents a docker container.
    This part is inspired by:
    http://techandtrains.com/2014/08/21/docker-container-as-mininet-host/
    We use the docker-py client library to control docker.
    """

    def __init__(self, name, dimage=None, dcmd=None, did=None, **kwargs):
        """
        Creates a Docker container as Mininet host.

        Resource limitations based on CFS scheduler:
        * cpu.cfs_quota_us: the total available run-time within a period (in microseconds)
        * cpu.cfs_period_us: the length of a period (in microseconds)
        (https://www.kernel.org/doc/Documentation/scheduler/sched-bwc.txt)

        Default Docker resource limitations:
        * cpu_shares: Relative amount of max. avail CPU for container
            (not a hard limit, e.g. if only one container is busy and the rest idle)
            e.g. usage: d1=4 d2=6 <=> 40% 60% CPU
        * cpuset_cpus: Bind containers to CPU 0 = cpu_1 ... n-1 = cpu_n (string: '0,2')
        * mem_limit: Memory limit (format: <number>[<unit>], where unit = b, k, m or g)
        * memswap_limit: Total limit = memory + swap

        All resource limits can be updated at runtime! Use:
        * updateCpuLimits(...)
        * updateMemoryLimits(...)
        """
        self.dimage = dimage
        self.dnameprefix = "mn"
        self.dcmd = dcmd if dcmd is not None else "/bin/bash"
        self.dc = None  # pointer to the dict containing 'Id' and 'Warnings' keys of the container
        self.dcinfo = None
        self.wintfs = {}
        self.wports = {}
        self.did = did  # Id of running container
        #  let's store our resource limits to have them available through the
        #  Mininet API later on
        defaults = { 'cpu_quota': -1,
                     'cpu_period': None,
                     'cpu_shares': None,
                     'cpuset_cpus': None,
                     'mem_limit': None,
                     'memswap_limit': None,
                     'environment': {},
                     'volumes': [],  # use ["/home/user1/:/mnt/vol2:rw"]
                     'tmpfs': [], # use ["/home/vol1/:size=3G,uid=1000"]
                     'network_mode': None,
                     'privileged': False,
                     'publish_all_ports': True,
                     'port_bindings': {},
                     'ports': [],
                     'dns': [],
                     'ipc_mode': None,
                     'devices': [],
                     'cap_add': ['net_admin'],  # we need this to allow mininet network setup
                     #'storage_opt': None,
                     'sysctls': {}
                     }
        defaults.update( kwargs )

        # keep resource in a dict for easy update during container lifetime
        self.resources = dict(
            cpu_quota=defaults['cpu_quota'],
            cpu_period=defaults['cpu_period'],
            cpu_shares=defaults['cpu_shares'],
            cpuset_cpus=defaults['cpuset_cpus'],
            mem_limit=defaults['mem_limit'],
            memswap_limit=defaults['memswap_limit']
        )

        self.volumes = defaults['volumes']
        self.tmpfs = defaults['tmpfs']
        self.environment = {} if defaults['environment'] is None else defaults['environment']
        # setting PS1 at "docker run" may break the python docker api (update_container hangs...)
        # self.environment.update({"PS1": chr(127)})  # CLI support
        self.network_mode = defaults['network_mode']
        self.privileged = defaults['privileged']
        self.publish_all_ports = defaults['publish_all_ports']
        self.port_bindings = defaults['port_bindings']
        self.dns = defaults['dns']
        self.ipc_mode = defaults['ipc_mode']
        self.devices = defaults['devices']
        self.cap_add = defaults['cap_add']
        self.sysctls = defaults['sysctls']
        #self.storage_opt = defaults['storage_opt']

        # setup docker client
        # self.dcli = docker.APIClient(base_url='unix://var/run/docker.sock')
        self.dcli = docker.from_env().api

        # pull image if it does not exist
        if self.dimage:
            self._check_image_exists(dimage, True)

        # for DEBUG
        debug("Created docker container object %s\n" % name)
        debug("image: %s\n" % str(self.dimage))
        debug("dcmd: %s\n" % str(self.dcmd))
        info("%s: kwargs %s\n" % (name, str(kwargs)))

        # creats host config for container
        # see: https://docker-py.readthedocs.org/en/latest/hostconfig/
        hc = self.dcli.create_host_config(
            network_mode=self.network_mode,
            privileged=self.privileged,
            binds=self.volumes,
            tmpfs=self.tmpfs,
            publish_all_ports=self.publish_all_ports,
            port_bindings=self.port_bindings,
            mem_limit=self.resources.get('mem_limit'),
            cpuset_cpus=self.resources.get('cpuset_cpus'),
            dns=self.dns,
            ipc_mode=self.ipc_mode,  # string
            devices=self.devices,  # see docker-py docu
            cap_add=self.cap_add,  # see docker-py docu
            sysctls=self.sysctls,  # see docker-py docu
            #storage_opt=self.storage_opt
        )

        if kwargs.get("rm", False):
            container_list = self.dcli.containers(all=True)
            for container in container_list:
                for container_name in container.get("Names", []):
                    if "%s.%s" % (self.dnameprefix, name) in container_name:
                        self.dcli.remove_container(container="%s.%s" % (self.dnameprefix, name), force=True)
                        break

        if self.did:
            self.dc = {'Id': '{}'.format(self.did), 'Warnings': []}
            self.existing_container = 1
        else:
            # create new docker container
            self.dc = self.dcli.create_container(
                name="%s.%s" % (self.dnameprefix, name),
                image=self.dimage,
                command=self.dcmd,
                entrypoint=list(),  # overwrite (will be executed manually at the end)
                stdin_open=True,  # keep container open
                tty=True,  # allocate pseudo tty
                environment=self.environment,
                #network_disabled=True,  # docker stats breaks if we disable the default network
                host_config=hc,
                ports=defaults['ports'],
                labels=['com.containernet'],
                volumes=[self._get_volume_mount_name(v) for v in self.volumes if self._get_volume_mount_name(v) is not None],
                hostname=name
            )

        # start the container
        self.dcli.start(self.dc)
        debug("Docker container %s started\n" % name)

        # fetch information about new container
        self.dcinfo = self.dcli.inspect_container(self.dc)
        self.did = self.dcinfo.get("Id")

        # call original Node.__init__
        Node.__init__(self, name, **kwargs)

        # let's initially set our resource limits
        self.update_resources(**self.resources)

    def start(self):
        # Containernet ignores the CMD field of the Dockerfile.
        # Lets try to load it here and manually execute it once the
        # container is started and configured by Containernet:
        cmd_field = self.get_cmd_field(self.dimage)
        entryp_field = self.get_entrypoint_field(self.dimage)
        if entryp_field is not None:
            if cmd_field is None:
                cmd_field = list()
            # clean up cmd_field
            try:
                cmd_field.remove(u'/bin/sh')
                cmd_field.remove(u'-c')
            except ValueError:
                pass
            # we just add the entryp. commands to the beginning:
            cmd_field = entryp_field + cmd_field
        if cmd_field is not None:
            cmd_field.append("> /dev/pts/0 2>&1")  # make output available to docker logs
            cmd_field.append("&")  # put to background (works, but not nice)
            info("{}: running CMD: {}\n".format(self.name, cmd_field))
            self.cmd(" ".join(cmd_field))

    def get_cmd_field(self, imagename):
        """
        Try to find the original CMD command of the Dockerfile
        by inspecting the Docker image.
        Returns list from CMD field if it is different from
        a single /bin/bash command which Containernet executes
        anyhow.
        """
        try:
            imgd = self.dcli.inspect_image(imagename)
            cmd = imgd.get("Config", {}).get("Cmd")
            assert isinstance(cmd, list)
            # filter the default case: a single "/bin/bash"
            if "/bin/bash" in cmd and len(cmd) == 1:
                return None
            return cmd
        except BaseException as ex:
            error("Error during image inspection of {}:{}"
                  .format(imagename, ex))
        return None

    def get_entrypoint_field(self, imagename):
        """
        Try to find the original ENTRYPOINT command of the Dockerfile
        by inspecting the Docker image.
        Returns list or None.
        """
        try:
            imgd = self.dcli.inspect_image(imagename)
            ep = imgd.get("Config", {}).get("Entrypoint")
            if isinstance(ep, list) and len(ep) < 1:
                return None
            return ep
        except BaseException as ex:
            error("Error during image inspection of {}:{}"
                  .format(imagename, ex))
        return None

    # Command support via shell process in namespace
    def startShell( self, *args, **kwargs ):
        "Start a shell process for running commands"
        if self.shell:
            error( "%s: shell is already running\n" % self.name )
            return
        # mnexec: (c)lose descriptors, (d)etach from tty,
        # (p)rint pid, and run in (n)amespace
        # opts = '-cd' if mnopts is None else mnopts
        # if self.inNamespace:
        #     opts += 'n'
        # bash -i: force interactive
        # -s: pass $* to shell, and make process easy to find in ps
        # prompt is set to sentinel chr( 127 )
        if hasattr(self, 'existing_container'):
            cmd = ['docker', 'exec', '-it', '%s' % self.did, 'env', 'PS1=' + chr(127),
                   'bash', '--norc', '-is', 'mininet:' + self.name]
        else:
            cmd = [ 'docker', 'exec', '-it',  '%s.%s' % ( self.dnameprefix, self.name ), 'env', 'PS1=' + chr( 127 ),
                    'bash', '--norc', '-is', 'mininet:' + self.name ]
        # Spawn a shell subprocess in a pseudo-tty, to disable buffering
        # in the subprocess and insulate it from signals (e.g. SIGINT)
        # received by the parent
        self.master, self.slave = pty.openpty()
        self.shell = self._popen( cmd, stdin=self.slave, stdout=self.slave, stderr=self.slave,
                                  close_fds=False )
        self.stdin = os.fdopen( self.master, 'r' )
        self.stdout = self.stdin
        self.pid = self._get_pid()
        self.pollOut = select.poll()
        self.pollOut.register( self.stdout )
        # Maintain mapping between file descriptors and nodes
        # This is useful for monitoring multiple nodes
        # using select.poll()
        self.outToNode[ self.stdout.fileno() ] = self
        self.inToNode[ self.stdin.fileno() ] = self
        self.execed = False
        self.lastCmd = None
        self.lastPid = None
        self.readbuf = ''
        # Wait for prompt
        while True:
            data = self.read( 1024 )
            if data[ -1 ] == chr( 127 ):
                break
            self.pollOut.poll()
        self.waiting = False
        # +m: disable job control notification
        self.cmd( 'unset HISTFILE; stty -echo; set +m' )

    def _get_volume_mount_name(self, volume_str):
        """ Helper to extract mount names from volume specification strings """
        parts = volume_str.split(":")
        if len(parts) < 3:
            return None
        return parts[1]

    def terminate( self ):
        """ Stop docker container """
        if not self._is_container_running():
            return
        try:
            if not hasattr(self, 'existing_container'):
                self.dcli.remove_container(self.dc, force=True, v=True)
        except docker.errors.APIError:
            warn("Warning: API error during container removal.\n")

        self.cleanup()

    def sendCmd( self, *args, **kwargs ):
        """Send a command, followed by a command to echo a sentinel,
           and return without waiting for the command to complete."""
        self._check_shell()
        if not self.shell:
            return
        Station.sendCmd( self, *args, **kwargs )

    def popen( self, *args, **kwargs ):
        """Return a Popen() object in node's namespace
           args: Popen() args, single list, or string
           kwargs: Popen() keyword args"""
        if not self._is_container_running():
            error( "ERROR: Can't connect to Container \'%s\'' for docker host \'%s\'!\n" % (self.did, self.name) )
            return
        mncmd = ["docker", "exec", "-t", "%s.%s" % (self.dnameprefix, self.name)]
        return Node.popen( self, *args, mncmd=mncmd, **kwargs )

    def cmd(self, *args, **kwargs ):
        """Send a command, wait for output, and return it.
           cmd: string"""
        verbose = kwargs.get( 'verbose', False )
        log = info if verbose else debug
        log( '*** %s : %s\n' % ( self.name, args ) )
        self.sendCmd( *args, **kwargs )
        return self.waitOutput( verbose )

    def _get_pid(self):
        state = self.dcinfo.get("State", None)
        if state:
            return state.get("Pid", -1)
        return -1

    def _check_shell(self):
        """Verify if shell is alive and
           try to restart if needed"""
        if self._is_container_running():
            if self.shell:
                self.shell.poll()
                if self.shell.returncode is not None:
                    debug("*** Shell died for docker host \'%s\'!\n" % self.name )
                    self.shell = None
                    debug("*** Restarting Shell of docker host \'%s\'!\n" % self.name )
                    self.startShell()
            else:
                debug("*** Restarting Shell of docker host \'%s\'!\n" % self.name )
                self.startShell()
        else:
            error( "ERROR: Can't connect to Container \'%s\'' for docker host \'%s\'!\n" % (self.did, self.name) )
            if self.shell:
                self.shell = None

    def _is_container_running(self):
        """Verify if container is alive"""
        container_list = self.dcli.containers(filters={"id": self.did, "status": "running"})
        if len(container_list) == 0:
            return False;
        return True

    def _check_image_exists(self, imagename, pullImage=False):
        # split tag from repository if a tag is specified
        if ":" in imagename:
            #If two :, then the first is to specify a port. Otherwise, it must be a tag
            slices = imagename.split(":")
            repo = ":".join(slices[0:-1])
            tag = slices[-1]
        else:
            repo = imagename
            tag = "latest"

        if self._image_exists(repo, tag):
            return True

        # image not found
        if pullImage:
            if self._pull_image(repo, tag):
                info('*** Download of "%s:%s" successful\n' % (repo, tag))
                return True
        # we couldn't find the image
        return False

    def _image_exists(self, repo, tag):
        """
        Checks if the repo:tag image exists locally
        :return: True if the image exists locally. Else false.
        """
        # filter by repository
        images = self.dcli.images(repo)
        imageName = "%s:%s" % (repo, tag)

        for image in images:
            if 'RepoTags' in image:
                if image['RepoTags'] is None:
                    return False
                if imageName in image['RepoTags']:
                    return True
        return False

    def _pull_image(self, repository, tag):
        """
        :return: True if pull was successful. Else false.
        """
        try:
            info('*** Image "%s:%s" not found. Trying to load the image. \n' % (repository, tag))
            info('*** This can take some minutes...\n')

            message = ""
            for line in self.dcli.pull(repository, tag, stream=True):
                # Collect output of the log for enhanced error feedback
                message = message + json.dumps(json.loads(line), indent=4)

        except:
            error('*** error: _pull_image: %s:%s failed.' % (repository, tag)
                  + message)
        if not self._image_exists(repository, tag):
            error('*** error: _pull_image: %s:%s failed.' % (repository, tag)
                  + message)
            return False
        return True

    def update_resources(self, **kwargs):
        """
        Update the container's resources using the docker.update function
        re-using the same parameters:
        Args:
           blkio_weight
           cpu_period, cpu_quota, cpu_shares
           cpuset_cpus
           cpuset_mems
           mem_limit
           mem_reservation
           memswap_limit
           kernel_memory
           restart_policy
        see https://docs.docker.com/engine/reference/commandline/update/
        or API docs: https://docker-py.readthedocs.io/en/stable/api.html#module-docker.api.container
        :return:
        """

        self.resources.update(kwargs)
        # filter out None values to avoid errors
        resources_filtered = {res:self.resources[res] for res in self.resources if self.resources[res] is not None}
        info("{1}: update resources {0}\n".format(resources_filtered, self.name))
        self.dcli.update_container(self.dc, **resources_filtered)

    def updateCpuLimit(self, cpu_quota=-1, cpu_period=-1, cpu_shares=-1, cores=None):
        """
        Update CPU resource limitations.
        This method allows to update resource limitations at runtime by bypassing the Docker API
        and directly manipulating the cgroup options.
        Args:
            cpu_quota: cfs quota us
            cpu_period: cfs period us
            cpu_shares: cpu shares
            cores: specifies which cores should be accessible for the container e.g. "0-2,16" represents
                Cores 0, 1, 2, and 16
        """
        # see https://www.kernel.org/doc/Documentation/scheduler/sched-bwc.txt

        # also negative values can be set for cpu_quota (uncontrained setting)
        # just check if value is a valid integer
        if isinstance(cpu_quota, int):
            self.resources['cpu_quota'] = self.cgroupSet("cfs_quota_us", cpu_quota)
        if cpu_period >= 0:
            self.resources['cpu_period'] = self.cgroupSet("cfs_period_us", cpu_period)
        if cpu_shares >= 0:
            self.resources['cpu_shares'] = self.cgroupSet("shares", cpu_shares)
        if cores:
            self.dcli.update_container(self.dc, cpuset_cpus=cores)
            # quota, period ad shares can also be set by this line. Usable for future work.

    def updateMemoryLimit(self, mem_limit=-1, memswap_limit=-1):
        """
        Update Memory resource limitations.
        This method allows to update resource limitations at runtime by bypassing the Docker API
        and directly manipulating the cgroup options.

        Args:
            mem_limit: memory limit in bytes
            memswap_limit: swap limit in bytes

        """
        # see https://www.kernel.org/doc/Documentation/scheduler/sched-bwc.txt
        if mem_limit >= 0:
            self.resources['mem_limit'] = self.cgroupSet("limit_in_bytes", mem_limit, resource="memory")
        if memswap_limit >= 0:
            self.resources['memswap_limit'] = self.cgroupSet("memsw.limit_in_bytes", memswap_limit, resource="memory")

    def cgroupSet(self, param, value, resource='cpu'):
        """
        Directly manipulate the resource settings of the Docker container's cgrpup.
        Args:
            param: parameter to set, e.g., cfs_quota_us
            value: value to set
            resource: resource name: cpu

        Returns: value that was set

        """
        cmd = 'cgset -r %s.%s=%s docker/%s' % (
            resource, param, value, self.did)
        debug(cmd + "\n")
        try:
            check_output(cmd, shell=True)
        except:
            error("Problem writing cgroup setting %r\n" % cmd)
            return
        nvalue = int(self.cgroupGet(param, resource))
        if nvalue != value:
            error('*** error: cgroupSet: %s set to %s instead of %s\n'
                  % (param, nvalue, value))
        return nvalue

    def cgroupGet(self, param, resource='cpu'):
        """
        Read cgroup values.
        Args:
            param: parameter to read, e.g., cfs_quota_us
            resource: resource name: cpu / memory

        Returns: value

        """
        cmd = 'cgget -r %s.%s docker/%s' % (
            resource, param, self.did)
        try:
            return int(check_output(cmd, shell=True).split()[-1])
        except:
            error("Problem reading cgroup info: %r\n" % cmd)
            return -1


class Switch(Node):
    """A Switch is a Node that is running (or has execed?)
       an OpenFlow switch."""

    portBase = 1  # Switches start with port 1 in OpenFlow
    dpidLen = 16  # digits in dpid passed to switch

    def __init__( self, name, dpid=None, opts='', listenPort=None, **params):
        """dpid: dpid hex string (or None to derive from name, e.g. s1 -> 1)
           opts: additional switch options
           listenPort: port to listen on for dpctl connections"""
        Node.__init__( self, name, **params )
        self.dpid = self.defaultDpid( dpid )
        self.opts = opts
        self.listenPort = listenPort
        if not self.inNamespace:
            self.controlIntf = Intf( 'lo', self, port=0 )

    def defaultDpid( self, dpid=None ):
        "Return correctly formatted dpid from dpid or switch name (s1 -> 1)"
        if dpid:
            # Remove any colons and make sure it's a good hex number
            dpid = dpid.replace( ':', '' )
            assert len( dpid ) <= self.dpidLen and int( dpid, 16 ) >= 0
        else:
            # Use hex of the first number in the switch name
            nums = re.findall( r'\d+', self.name )
            if nums:
                dpid = hex( int( nums[ 0 ] ) )[ 2: ]
            else:
                self.terminate()  # Python 3.6 crash workaround
                raise Exception( 'Unable to derive default datapath ID - '
                                 'please either specify a dpid or use a '
                                 'canonical switch name such as s23.' )
        return '0' * ( self.dpidLen - len( dpid ) ) + dpid

    def defaultIntf( self ):
        "Return control interface"
        if self.controlIntf:
            return self.controlIntf
        else:
            return Node.defaultIntf( self )

    def sendCmd( self, *cmd, **kwargs ):
        """Send command to Node.
           cmd: string"""
        kwargs.setdefault( 'printPid', False )
        if not self.execed:
            return Node.sendCmd( self, *cmd, **kwargs )
        else:
            error( '*** Error: %s has execed and cannot accept commands' %
                   self.name )

    def connected( self ):
        "Is the switch connected to a controller? (override this method)"
        # Assume that we are connected by default to whatever we need to
        # be connected to. This should be overridden by any OpenFlow
        # switch, but not by a standalone bridge.
        debug( 'Assuming', repr( self ), 'is connected to a controller\n' )
        return True

    def stop( self, deleteIntfs=True ):
        """Stop switch
           deleteIntfs: delete interfaces? (True)"""
        if deleteIntfs:
            self.deleteIntfs()
        self.terminate()

    def __repr__( self ):
        "More informative string representation"
        intfs = ( ','.join( [ '%s:%s' % ( i.name, i.IP() )
                              for i in self.intfList() ] ) )
        return '<%s %s: %s pid=%s> ' % (
            self.__class__.__name__, self.name, intfs, self.pid )


class UserSwitch( Switch ):
    "User-space switch."

    dpidLen = 12

    def __init__( self, name, dpopts='--no-slicing', **kwargs ):
        """Init.
           name: name for the switch
           dpopts: additional arguments to ofdatapath (--no-slicing)"""
        Switch.__init__( self, name, **kwargs )
        pathCheck( 'ofdatapath', 'ofprotocol',
                   moduleName='the OpenFlow reference user switch' +
                              '(openflow.org)' )
        if self.listenPort:
            self.opts += ' --listen=ptcp:%i ' % self.listenPort
        else:
            self.opts += ' --listen=punix:/tmp/%s.listen' % self.name
        self.dpopts = dpopts

    @classmethod
    def setup( cls ):
        "Ensure any dependencies are loaded; if not, try to load them."
        if not os.path.exists( '/dev/net/tun' ):
            moduleDeps( add=TUN )

    def dpctl( self, *args ):
        "Run dpctl command"
        listenAddr = None
        if not self.listenPort:
            listenAddr = 'unix:/tmp/%s.listen' % self.name
        else:
            listenAddr = 'tcp:127.0.0.1:%i' % self.listenPort
        return self.cmd( 'dpctl ' + ' '.join( args ) +
                         ' ' + listenAddr )

    def connected( self ):
        "Is the switch connected to a controller?"
        status = self.dpctl( 'status' )
        return ( 'remote.is-connected=true' in status and
                 'local.is-connected=true' in status )

    @staticmethod
    def TCReapply( intf ):
        """Unfortunately user switch and Mininet are fighting
           over tc queuing disciplines. To resolve the conflict,
           we re-create the user switch's configuration, but as a
           leaf of the TCIntf-created configuration."""
        if isinstance( intf, TCIntf ):
            ifspeed = 10000000000  # 10 Gbps
            minspeed = ifspeed * 0.001

            res = intf.config( **intf.params )

            if res is None:  # link may not have TC parameters
                return

            # Re-add qdisc, root, and default classes user switch created, but
            # with new parent, as setup by Mininet's TCIntf
            parent = res['parent']
            intf.tc( "%s qdisc add dev %s " + parent +
                     " handle 1: htb default 0xfffe" )
            intf.tc( "%s class add dev %s classid 1:0xffff parent 1: htb rate "
                     + str(ifspeed) )
            intf.tc( "%s class add dev %s classid 1:0xfffe parent 1:0xffff " +
                     "htb rate " + str(minspeed) + " ceil " + str(ifspeed) )

    def start( self, controllers ):
        """Start OpenFlow reference user datapath.
           Log to /tmp/sN-{ofd,ofp}.log.
           controllers: list of controller objects"""
        # Add controllers
        clist = ','.join( [ 'tcp:%s:%d' % ( c.IP(), c.port )
                            for c in controllers ] )
        ofdlog = '/tmp/' + self.name + '-ofd.log'
        ofplog = '/tmp/' + self.name + '-ofp.log'
        intfs = [ str( i ) for i in self.intfList() if not i.IP() ]
        self.cmd( 'ofdatapath -i ' + ','.join( intfs ) +
                  ' punix:/tmp/' + self.name + ' -d %s ' % self.dpid +
                  self.dpopts +
                  ' 1> ' + ofdlog + ' 2> ' + ofdlog + ' &' )
        self.cmd( 'ofprotocol unix:/tmp/' + self.name +
                  ' ' + clist +
                  ' --fail=closed ' + self.opts +
                  ' 1> ' + ofplog + ' 2>' + ofplog + ' &' )
        if "no-slicing" not in self.dpopts:
            # Only TCReapply if slicing is enable
            sleep(1)  # Allow ofdatapath to start before re-arranging qdisc's
            for intf in self.intfList():
                if not intf.IP():
                    self.TCReapply( intf )

    def stop( self, deleteIntfs=True ):
        """Stop OpenFlow reference user datapath.
           deleteIntfs: delete interfaces? (True)"""
        self.cmd( 'kill %ofdatapath' )
        self.cmd( 'kill %ofprotocol' )
        super( UserSwitch, self ).stop( deleteIntfs )


class OVSSwitch( Switch ):
    "Open vSwitch switch. Depends on ovs-vsctl."

    def __init__( self, name, failMode='secure', datapath='kernel',
                  inband=False, protocols=None,
                  reconnectms=1000, stp=False, batch=False, **params ):
        """name: name for switch
           failMode: controller loss behavior (secure|standalone)
           datapath: userspace or kernel mode (kernel|user)
           inband: use in-band control (False)
           protocols: use specific OpenFlow version(s) (e.g. OpenFlow13)
                      Unspecified (or old OVS version) uses OVS default
           reconnectms: max reconnect timeout in ms (0/None for default)
           stp: enable STP (False, requires failMode=standalone)
           batch: enable batch startup (False)"""
        Switch.__init__( self, name, **params )
        self.failMode = failMode
        self.datapath = datapath
        self.inband = inband
        self.protocols = protocols
        self.reconnectms = reconnectms
        self.stp = stp
        self._uuids = []  # controller UUIDs
        self.batch = batch
        self.commands = []  # saved commands for batch startup

        # add a prefix to the name of the deployed switch, to find it back easier later
        prefix = params.get('prefix', '')
        self.deployed_name = prefix + name


    @classmethod
    def setup( cls ):
        "Make sure Open vSwitch is installed and working"
        pathCheck( 'ovs-vsctl',
                   moduleName='Open vSwitch (openvswitch.org)')
        # This should no longer be needed, and it breaks
        # with OVS 1.7 which has renamed the kernel module:
        #  moduleDeps( subtract=OF_KMOD, add=OVS_KMOD )
        out, err, exitcode = errRun( 'ovs-vsctl -t 1 show' )
        if exitcode:
            error( out + err +
                   'ovs-vsctl exited with code %d\n' % exitcode +
                   '*** Error connecting to ovs-db with ovs-vsctl\n'
                   'Make sure that Open vSwitch is installed, '
                   'that ovsdb-server is running, and that\n'
                   '"ovs-vsctl show" works correctly.\n'
                   'You may wish to try '
                   '"service openvswitch-switch start".\n' )
            exit( 1 )
        version = quietRun( 'ovs-vsctl --version' )
        cls.OVSVersion = findall( r'\d+\.\d+', version )[ 0 ]

    @classmethod
    def isOldOVS( cls ):
        "Is OVS ersion < 1.10?"
        return ( StrictVersion( cls.OVSVersion ) <
                 StrictVersion( '1.10' ) )

    def dpctl( self, *args ):
        "Run ovs-ofctl command"
        return self.cmd( 'ovs-ofctl', args[ 0 ], self.deployed_name, *args[ 1: ] )

    def vsctl( self, *args, **kwargs ):
        "Run ovs-vsctl command (or queue for later execution)"
        if self.batch:
            cmd = ' '.join( str( arg ).strip() for arg in args )
            self.commands.append( cmd )
        else:
            return self.cmd( 'ovs-vsctl', *args, **kwargs )

    @staticmethod
    def TCReapply( intf ):
        """Unfortunately OVS and Mininet are fighting
           over tc queuing disciplines. As a quick hack/
           workaround, we clear OVS's and reapply our own."""
        if isinstance( intf, TCIntf ):
            intf.config( **intf.params )

    def attach( self, intf ):
        "Connect a data port"
        self.vsctl( 'add-port', self.deployed_name, intf )
        self.cmd( 'ifconfig', intf, 'up' )
        self.TCReapply( intf )

    def attachInternalIntf(self, intf_name, net):
        """Add an interface of type:internal to the ovs switch
           and add routing entry to host"""
        self.vsctl('add-port', self.deployed_name, intf_name, '--', 'set',
                   ' interface', intf_name, 'type=internal')
        int_intf = Intf(intf_name, node=self.deployed_name)
        #self.addIntf(int_intf, moveIntfFn=None)
        self.cmd('ip route add', net, 'dev', intf_name)

        return self.nameToIntf[intf_name]

    def detach( self, intf ):
        "Disconnect a data port"
        self.vsctl( 'del-port', self.deployed_name, intf )

    def controllerUUIDs( self, update=False ):
        """Return ovsdb UUIDs for our controllers
           update: update cached value"""
        if not self._uuids or update:
            controllers = self.cmd( 'ovs-vsctl -- get Bridge', self.deployed_name,
                                    'Controller' ).strip()
            if controllers.startswith( '[' ) and controllers.endswith( ']' ):
                controllers = controllers[ 1 : -1 ]
                if controllers:
                    self._uuids = [ c.strip()
                                    for c in controllers.split( ',' ) ]
        return self._uuids

    def connected( self ):
        "Are we connected to at least one of our controllers?"
        for uuid in self.controllerUUIDs():
            if 'true' in self.vsctl( '-- get Controller',
                                     uuid, 'is_connected' ):
                return True
        return self.failMode == 'standalone'

    def intfOpts( self, intf ):
        "Return OVS interface options for intf"
        opts = ''
        if not self.isOldOVS():
            # ofport_request is not supported on old OVS
            opts += ' ofport_request=%s' % self.ports[ intf ]
            # Patch ports don't work well with old OVS
            if isinstance( intf, OVSIntf ):
                intf1, intf2 = intf.link.intf1, intf.link.intf2
                peer = intf1 if intf1 != intf else intf2
                opts += ' type=patch options:peer=%s' % peer
        return '' if not opts else ' -- set Interface %s' % intf + opts

    def bridgeOpts( self ):
        "Return OVS bridge options"
        opts = ( ' other_config:datapath-id=%s' % self.dpid +
                 ' fail_mode=%s' % self.failMode )
        if not self.inband:
            opts += ' other-config:disable-in-band=true'
        if self.datapath == 'user':
            opts += ' datapath_type=netdev'
        if self.protocols and not self.isOldOVS():
            opts += ' protocols=%s' % self.protocols
        if self.stp and self.failMode == 'standalone':
            opts += ' stp_enable=true'
        opts += ' other-config:dp-desc=%s' % self.name
        return opts

    def start( self, controllers ):
        "Start up a new OVS OpenFlow switch using ovs-vsctl"
        if self.inNamespace:
            raise Exception(
                'OVS kernel switch does not work in a namespace' )
        int( self.dpid, 16 )  # DPID must be a hex string
        # Command to add interfaces
        intfs = ''.join( ' -- add-port %s %s' % ( self.deployed_name, intf ) +
                         self.intfOpts( intf )
                         for intf in self.intfList()
                         if self.ports[ intf ] and not intf.IP() )
        # Command to create controller entries
        clist = [ ( self.deployed_name + c.name, '%s:%s:%d' %
                  ( c.protocol, c.IP(), c.port ) )
                  for c in controllers ]
        if self.listenPort:
            clist.append( ( self.deployed_name + '-listen',
                            'ptcp:%s' % self.listenPort ) )
        ccmd = '-- --id=@%s create Controller target=\\"%s\\"'
        if self.reconnectms:
            ccmd += ' max_backoff=%d' % self.reconnectms
        cargs = ' '.join( ccmd % ( name, target )
                          for name, target in clist )
        # Controller ID list
        cids = ','.join( '@%s' % name for name, _target in clist )
        # Try to delete any existing bridges with the same name
        if not self.isOldOVS():
            cargs += ' -- --if-exists del-br %s' % self.deployed_name
        # One ovs-vsctl command to rule them all!
        self.vsctl( cargs +
                    ' -- add-br %s' % self.deployed_name +
                    ' -- set bridge %s controller=[%s]' % ( self.deployed_name, cids  ) +
                    self.bridgeOpts() +
                    intfs )
        # If necessary, restore TC config overwritten by OVS
        if not self.batch:
            for intf in self.intfList():
                self.TCReapply( intf )

    # This should be ~ int( quietRun( 'getconf ARG_MAX' ) ),
    # but the real limit seems to be much lower
    argmax = 128000

    @classmethod
    def batchStartup( cls, switches, run=errRun ):
        """Batch startup for OVS
           switches: switches to start up
           run: function to run commands (errRun)"""
        info( '...' )
        cmds = 'ovs-vsctl'
        for switch in switches:
            if switch.isOldOVS():
                # Ideally we'd optimize this also
                run( 'ovs-vsctl del-br %s' % switch )
            for cmd in switch.commands:
                cmd = cmd.strip()
                # Don't exceed ARG_MAX
                if len( cmds ) + len( cmd ) >= cls.argmax:
                    run( cmds, shell=True )
                    cmds = 'ovs-vsctl'
                cmds += ' ' + cmd
                switch.cmds = []
                switch.batch = False
        if cmds:
            run( cmds, shell=True )
        # Reapply link config if necessary...
        for switch in switches:
            for intf in switch.intfs.values():
                if isinstance( intf, TCIntf ):
                    intf.config( **intf.params )
        return switches

    def stop( self, deleteIntfs=True ):
        """Terminate OVS switch.
           deleteIntfs: delete interfaces? (True)"""
        self.cmd( 'ovs-vsctl del-br', self.deployed_name )
        if self.datapath == 'user':
            self.cmd( 'ip link del', self.deployed_name )
        super( OVSSwitch, self ).stop( deleteIntfs )

    @classmethod
    def batchShutdown( cls, switches, run=errRun ):
        "Shut down a list of OVS switches"
        delcmd = 'del-br %s'
        if switches and not switches[ 0 ].isOldOVS():
            delcmd = '--if-exists ' + delcmd
        # First, delete them all from ovsdb
        run( 'ovs-vsctl ' +
             ' -- '.join( delcmd % s.deployed_name for s in switches ), shell=True )
        # Next, shut down all of the processes
        pids = ' '.join( str( switch.pid ) for switch in switches )

        success = False
        while not success:
            try:
                run( 'kill -HUP ' + pids )
                success = True
            except select.error as e:
                # retry on interrupt
                if e[0] != errno.EINTR:
                    raise
        for switch in switches:
            switch.terminate()
            switch.shell = None
        return switches


OVSKernelSwitch = OVSSwitch



from .deploy_lft import DeployLFT
from time import time, sleep
from .deploy_mininet import DeployMininet
from pandas import DataFrame, read_csv, concat
from matplotlib import pyplot as plt
from subprocess import run
from threading import Thread
from os.path import isfile
from .constants import *


def barPlot(dataframe, title):
    [plt.bar(dataframe.index, dataframe[column], label=column) for column in dataframe.columns]
    plt.title(title)
    plt.xlabel("Nodes")
    plt.legend()
    plt.show()


def cleanupContainers():
    out = run('docker ps -qa', shell=True, capture_output=True)
    containerIds = out.stdout.decode().split()
    [run(f'docker rm -f {id}', shell=True) for id in containerIds]


def saveFile(dataframe, filename):
    if isfile(filename):
        dataframe = concat([dataframe, read_csv(filename)], axis=0)
    dataframe.to_csv(filename, index=False)


awaitStabilizeMemoryTime = 5
maxMem = 0
continueThread = True
result = 0
def measureMemory(startMemoryUsage):
    global result, maxMem, continueThread
    maxMem = startMemoryUsage
    while(continueThread):
        aux = getCurrentMemoryUsage()
        if aux > maxMem:
            maxMem = aux
            result = maxMem - startMemoryUsage
        sleep(1)


def getCurrentMemoryUsage():
    command = 'free | grep Mem | grep -oP \'^\D*\d+\D*\K\d+\''
    out = run(command, shell=True, capture_output=True, text=True)
    return int(out.stdout)


replicas = 30
sizes = [1, 4, 16, 64, 256]
coolDownTime = 20
cleanupContainers()


# Measure deployment and Undeployment time of LFT
deployLftDf = DataFrame(columns = sizes)
undeployLftDf = DataFrame(columns = sizes)
dlft = DeployLFT()
for i in range(replicas):
    print(f'LFT Deployment and Undeployment Assessment: Replica {i+1}')
    lftDeployTime = []
    lftUndeployTime = []
    for size in sizes:
        try:
            print(f'Deploying {size} node(s)')
            start = time()
            dlft.deploy(size)
            lftDeployTime.append(time() - start)
            print(f'LFT Deployment Time: {lftDeployTime}')
            dlft.getReferences(size)
            sleep(coolDownTime)
            start = time()
            dlft.undeploy()
            end = time()
            lftUndeployTime.append(time() - start)
            print(f'LFT Undeployment Time: {lftUndeployTime}')
            sleep(coolDownTime)
        except Exception as ex:
            print(f"Caught an exception. {ex}")
            dlft.getReferences(sizes)
            cleanupContainers()
            continue
    print(f'LFT Deployment times for replica {i+1} were {lftDeployTime}')
    print(f'LFT Undeployment times for replica {i+1} were {lftUndeployTime}')
    deployLftDf.loc[i] = lftDeployTime
    undeployLftDf.loc[i] = lftUndeployTime


cleanupContainers()
saveFile(deployLftDf, f'{RESULTS_PATH}deployLftTime.csv')
saveFile(undeployLftDf, f'{RESULTS_PATH}undeployLftTime.csv')
#barPlot(deployLftDf, "LFT deployment time")
#barPlot(undeployLftDf, "LFT undeployment time")



# Measure Memory consumption of LFT
awaitStabilizeMemoryTime = 5

deployMemLftDf = DataFrame(columns = sizes)
for i in range(replicas):
    print(f'LFT Memory Consumption Assessment: Replica {i+1}')
    lftDeployMem = []
    for size in sizes:
        try:
            print(f'Deploying {size} node(s)')
            maxMem = 0
            continueThread = True
            startMemoryUsage = getCurrentMemoryUsage()
            t = Thread(target=measureMemory, args=(startMemoryUsage,))
            t.start()
            sleep(awaitStabilizeMemoryTime)
            dlft.deploy(size)
            sleep(awaitStabilizeMemoryTime)
            continueThread = False
            t.join()
            lftDeployMem.append(result)
            print(f'LFT Deployment Max Memory Consumption: {lftDeployMem}')
            dlft.getReferences(size)
            dlft.undeploy()
            sleep(coolDownTime)
        except Exception as ex:
            print(f"Caught an exception. {ex}")
            dlft.getReferences(sizes)
            cleanupContainers()
            continue
    print(f'LFT Deployment Memory Consumption for replica {i+1} were {lftDeployMem}')
    deployMemLftDf.loc[i] = lftDeployMem


cleanupContainers()
saveFile(deployMemLftDf, f'{RESULTS_PATH}deployLftMem.csv')



# Measure deployment and Undeployment time of Mininet
deployMnDf = DataFrame(columns = sizes)
undeployMnDf = DataFrame(columns = sizes)
for i in range(replicas):
    print(f'Mininet-WiFi Deployment and Undeployment Assessment: Replica {i+1}')
    mnDeployTime = []
    mnUndeployTime = []
    dmn = DeployMininet()
    for size in sizes:
        try:
            print(f'Deploying {size} node(s)')
            start = time()
            dmn.deploy(size)
            mnDeployTime.append(time() - start)
            print(f'Mininet-WiFi Deployment Time: {mnDeployTime}')
            sleep(coolDownTime)
            start = time()
            dmn.undeploy()
            mnUndeployTime.append(time() - start)
            print(f'Mininet-WiFi Undeployment Time: {mnUndeployTime}')
            sleep(coolDownTime)
        except Exception as ex:
            print(f"Caught an exception. {ex}")
            cleanupContainers()
            continue
    print(f'Mininet Containernet Deployment times for replica {i+1} were {mnDeployTime}')
    print(f'Mininet Containernet Undeployment times for replica {i+1} were {mnUndeployTime}')
    deployMnDf.loc[i] = mnDeployTime
    undeployMnDf.loc[i] = mnUndeployTime


cleanupContainers()
saveFile(deployMnDf, f'{RESULTS_PATH}deployMnTime.csv')
saveFile(undeployMnDf, f'{RESULTS_PATH}undeployMnTime.csv')
#barPlot(deployMnDf, "ContainerNet deployment time")
#barPlot(undeployLftDf, "ContainerNet undeployment time")



# Measure deployment memory of Mininet
deployMemMnDf = DataFrame(columns = sizes)
for i in range(replicas):
    print(f'Mininet-Wifi Memory Consumption Assessment: Replica {i+1}')
    mnDeployMem = []
    dmn = DeployMininet()
    for size in sizes:
        try:
            print(f'Deploying {size} node(s)')
            maxMem = 0
            result = 0
            continueThread = True
            startMemoryUsage = getCurrentMemoryUsage()
            t = Thread(target=measureMemory, args=(startMemoryUsage,))
            t.start()
            sleep(awaitStabilizeMemoryTime)
            dmn.deploy(size)
            sleep(awaitStabilizeMemoryTime)
            continueThread = False
            t.join()
            mnDeployMem.append(result)
            print(f'Mininet-WiFi Deployment Max Memory Consumption: {mnDeployMem}')
            dmn.undeploy()
            sleep(coolDownTime)
        except Exception as ex:
            print(f"Caught an exception. {ex}")
            cleanupContainers()
            continue
    print(f'Mininet-WiFi Deployment Memory Consumption for replica {i+1} were {mnDeployMem}')
    deployMemMnDf.loc[i] = mnDeployMem


cleanupContainers()
saveFile(deployMemMnDf, f'{RESULTS_PATH}deployMnMem.csv')



# Copyright (C) 2022 Alexandre Mitsuru Kaihara
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .node import Node
from .exceptions import NodeInstantiationFailed


class CICFlowMeter(Node): 
    # Brief: Instantiate a CICFlowMeter class, where it can be defined to capture flow data of each interface added
    # Params:
    # Return:
    #   None
    def __init__(self, name: str, hostPath='', containerPath=''):
        super().__init__(name)
        self.__mount = False
        if hostPath != '' and containerPath != '':
            self.__hostPath = hostPath
            self.__containerPath = containerPath
            self.__mount = True
        else: 
            raise Exception(f"Invalid hostPath and containerPath mount point on {self.getNodeName()}. hostPath and containerPath cannot be null")

    # Brief: Instantiate an OpenvSwitch switch container
    # Params:
    # Return:
    #   None
    def instantiate(self, image='alexandremitsurukaihara/lst2.0:cicflowmeter') -> None:
        mount = ''
        if self.__mount: mount = f'-v {self.__hostPath}:{self.__containerPath}'
        super().instantiate(dockerCommand=f"docker run -d --network=none --privileged {mount} --name={self.getNodeName()} {image}")
        
    # Brief: Set up the tshark to sniff all the packets into pcap files
    # Params:
    #   List<Node> nodes: References of the nodes connected to this switch to sniff packets
    #   boolean sniffAll: If sniff all is set  
    # Return:
    def analyze(self, pcapPath: str, destPath) -> None:
        self.run(f'./TCPDUMP_and_CICFlowMeter-master/convert_pcap_csv.sh {pcapPath}')
        self.run('find /TCPDUMP_and_CICFlowMeter-master/csv -type f -exec mv {}' + f' {destPath} \\;')



from profissa_lft.host import Host
from experiment.constants import *
import subprocess

class EmuPhyWired:
    def __init__(self):
        self.h1 = Host("h1")
    def setup(self):    
        self.h1.instantiate(dockerImage=PERFSONAR_TESTPOINT_UBUNTU, runCommand=USR_SBIN_INIT_COMMAND)
        self.h1.connectToInternet("10.0.0.1", 30, "h1host", "hosth1")
        self.h1.setIp(EMU_PHY_WIRED_H1_IP, 24, "h1host")
        self.h1.setDefaultGateway("10.0.0.1", "h1host")
        self.h1.setHost('10.0.0.1')
        #self.h1.setInterfaceProperties("h1host", "935mbit", "0.3ms", "0.15ms")
        self.h1.setMtuSize('h1host', 9000)
        subprocess.run("ip link set dev hosth1 mtu 9000", shell=True)
    def tearDown(self):
        self.h1.delete()



# Copyright (C) 2022 Alexandre Mitsuru Kaihara
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import subprocess
from node import Node
from exceptions import NodeInstantiationFailed


class Controller(Node):
    def __init__(self, nodeName: str) -> None:
        super().__init__(nodeName)
        self.__process = 0

    # Brief: Instantiate a controller container
    # Params:
    # Return:
    #   None
    def instantiate(self, dockerImage='alexandremitsurukaihara/lst2.0:ryucontroller', dockerCommand='') -> None:
        super().instantiate(dockerImage=dockerImage, dockerCommand=dockerCommand)

    # Brief: Instantiate a controller container, the ip and port will only be used if the command parameter is an empty list. In this case, will be instantiated a Ryu controller.
    # Params:
    #   String ip: Ip address to which the controller will be listening to
    #   int port: Number of the port that the controller will be listening to
    #   List<String> command: List of commands to execute in the controller to instantiate the controller
    # Return:
    #   None
    def initController(self, ip:str, port: int, command=[]):
        try:
            if len(command) == 0:
                subprocess.run(f"docker exec {self.getNodeName()} ryu-manager --ofp-listen-host={ip} --ofp-tcp-listen-port={port} /home/controller.py > /dev/null 2>&1 &", shell=True)
            else:
                for c in command: subprocess.run(c, shell=True)
        except Exception as ex:
            logging.error(f"Error while setting up controller {self.getNodeName()} in {ip}/{port}: {str(ex)}")
            raise Exception(f"Error while setting up controller {self.getNodeName()} in {ip}/{port}: {str(ex)}")

    def instantiate_local(self, controllerIp, controllerPort):
        process = self.__getProcess()
        if process == 0:
            try:
                self.__process = subprocess.Popen(f"ryu-manager --ofp-listen-host={controllerIp} --ofp-tcp-listen-port={controllerPort} controller.py > controller.log", shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            except Exception as ex:
                logging.error(f"Error while creating the switch {self.getNodeName()}: {str(ex)}")
                raise NodeInstantiationFailed(f"Error while creating the switch {self.getNodeName()}: {str(ex)}")
        else:
            logging.error(f"Controller {self.getNodeName()} already instantiated")
            raise Exception(f"Controller {self.getNodeName()} already instantiated")
        
    def delete_local(self):
        process = self.__getProcess()
        if process != 0:
            try:
                self.__process.kill()
                _, stderr = self.__process.communicate()
            except Exception as ex:
                logging.error(f"Error while deleting the switch {self.getNodeName()}: {str(ex)}\nThreads error: {stderr}")
                raise NodeInstantiationFailed(f"Error while deleting the switch {self.getNodeName()}: {str(ex)}\nThreads error: {stderr}")
        else:
            logging.error(f"Can't delete {self.getNodeName()}. {self.getNodeName()} was not instantiated.")
            raise Exception(f"Can't delete {self.getNodeName()}. {self.getNodeName()} was not instantiated.")
        
    def __getProcess(self):
        return self.__process


from profissa_lft.perfsonar import Perfsonar
from time import sleep
from experiment.constants import *

class EmuEmuWired:
    def __init__(self):
        self.h1 = Perfsonar('h1')
        self.h2 = Perfsonar('h2')

    def setup(self):
        self.h1.instantiate(PERFSONAR_TESTPOINT_UBUNTU, runCommand=USR_SBIN_INIT_COMMAND)
        self.h2.instantiate(PERFSONAR_TESTPOINT_UBUNTU, runCommand=USR_SBIN_INIT_COMMAND)
        
        self.h1.connect(self.h2, "h1h2", "h2h1")
        self.h1.setIp(EMU_EMU_WIRED_H1_IP, 24, "h1h2")
        self.h2.setIp(EMU_EMU_WIRED_H2_IP, 24, "h2h1")
        
        self.h1.connectToInternet(EMU_EMU_WIRED_HOST_IP, 24, "h1host", "hosth1")
        self.h1.setIp("192.0.0.2", 24, "h1host")

        self.h1.enableForwarding("h1host", "h1h2")
        self.h1.setDefaultGateway(EMU_EMU_WIRED_HOST_IP, "h1host")
        self.h2.setDefaultGateway(EMU_EMU_WIRED_H1_IP, "h2h1")

        self.h1.addRouteOnHost('10.0.0.0', 24, 'hosth1', '192.0.0.2')

        self.h1.setHost(EMU_EMU_WIRED_H1_IP)
        self.h2.setHost(EMU_EMU_WIRED_H2_IP)

        self.h1.readLimitFile()
        self.h2.readLimitFile()

        self.h1.addRouteException("192.0.0.0", 24)
        self.h2.addRouteException("192.0.0.0", 24)

        self.h1.saveLimitFile()
        self.h2.saveLimitFile()

        self.__startPerfsonarServices()

        self.h1.setInterfaceProperties("h1h2", "935mbit", "0.3ms", "0.15ms")
        self.h2.setInterfaceProperties("h2h1", "935mbit", "0.3ms", "0.15ms")
        
    def __startPerfsonarServices(self):
         self.h1.run("service pscheduler-runner start && service pscheduler-ticker start && service pscheduler-scheduler start && service pscheduler-archiver start")
         sleep(5)
         self.h2.run("service pscheduler-runner start && service pscheduler-ticker start && service pscheduler-scheduler start && service pscheduler-archiver start")


    def tearDown(self):
        self.h1.delete()
        self.h2.delete()
